#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <cstring>
#include <math.h>
#include "ft_stuff.h"
#include "FTMFile.h"

#define MAGIC_NUMBER	110250

#define READ_DWORD(dst)			\
{								\
	memcpy(dst, &buf[dp], 4);	\
	dp += 4;					\
}
							

//TODO: Allocate stuff dynamically ;)

byte saw_wave[32]={0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15};
byte dummy_seq[32]={15,15,14,14,13,13,12,12,11,11,10,10,9,9,8,8,7,7,6,6,5,5,4,4,3,3,2,2,1,1,0,0};

stSequence::stSequence()
{
	memset(this,0,sizeof(stSequence));
	loop = -1;
	release = -1;
}

Instrument::Instrument()
{
	memset(this, 0, sizeof(Instrument));
	type = 1;
	seq_count = SEQ_COUNT;
	wave_size = 32;
	wave_count = 1;
	memcpy(wave_data, saw_wave, 32);
	name_length = 0xE;
	memcpy(name,"New instrument",0xE);
}

FTItem::FTItem()
{
	memset(this,0,sizeof(FTItem));
	instrument=64;
	volume=16;
}

FTPattern::FTPattern()
{
	memset(this,0,sizeof(FTPattern));
}

CFTMFile::CFTMFile()
{
	Reset();
}

void CFTMFile::Reset()
{
	int i;
	memset(this,0,sizeof(CFTMFile));
	strcpy(file_id,"FamiTracker Module");
	file_version = 0x440;

	channels = 5;
	v_style = 1;
	highlight1 = 4;
	highlight2 = 16;
	s_split = 32;

	for(i=0;i<32;i++){
		ch_id[i] = i;
		ch_fx[i] = 0;
	}
	
	n_frames = 1;
	speed = 6;
	tempo = 150;
	p_length = 64;
}

bool CFTMFile::LoadFile(char* filename)
{
	unsigned char* buf;
	FILE* fp_src;
	char block_id[16];
	int src_size;
	int total_instruments;
	int inst_index;
	int inst_type;
	int patterns_end;
	int namelen;
	int seqlen;
	int block_version;
	int block_size;
	int dp;
	int i, j;
	
	Reset();
	
	dp = 0;
	
	fp_src = fopen(filename,"rb"); if(fp_src==NULL) {printf("Error! Source FTM file not found. :(\n"); return false;}
	fseek(fp_src, 0, SEEK_END); src_size = ftell(fp_src); fseek(fp_src, 0, SEEK_SET); //Get file size
	buf = new unsigned char[src_size];
	fread(buf, src_size, 1, fp_src);
	fclose(fp_src);
	
	memcpy(file_id, &buf[dp], 18); dp += 18;
	memcpy(&file_version, &buf[dp], 4); dp += 4;
	
	printf("PARAMS @ offset: 0x%X\n", dp);
	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;
	expansion = buf[dp++];
	memcpy(&channels, &buf[dp], 4); dp += 4;
	memcpy(&machine, &buf[dp], 4); dp += 4;
	memcpy(&e_speed, &buf[dp], 4); dp += 4;
	memcpy(&v_style, &buf[dp], 4); dp += 4;
	memcpy(&highlight1, &buf[dp], 4); dp += 4;
	memcpy(&highlight2, &buf[dp], 4); dp += 4;
	if(expansion&0x10){
		memcpy(&N163_channels, &buf[dp], 4);
		dp += 4;
	}
	memcpy(&s_split, &buf[dp], 4); dp += 4;

	printf("INFO @ offset: 0x%X\n", dp);
	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;
	memcpy(author, &buf[dp], 32); dp += 32;
	memcpy(title, &buf[dp], 32); dp += 32;
	memcpy(ccccc, &buf[dp], 32); dp += 32;
	
	printf("HEADER @ offset: 0x%X\n", dp);
	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;
	n_tracks = buf[dp++];
	while(buf[dp++]); //Skip null-terminated string
	for(i=0;i<channels;i++){
		ch_id[i] = buf[dp++];
		ch_fx[i] = buf[dp++];	
	}

	printf("INSTRUMENTS @ offset: 0x%X\n", dp);
	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;
	memcpy(&n_instruments, &buf[dp], 4); dp += 4;
	
	for(i=0;i<n_instruments;i++){
		int c_index;
		int c_type;
		memcpy(&c_index, &buf[dp], 4); dp += 4;
		c_type = buf[dp++];
		
		Inst[i].index = c_index;
		Inst[i].type = c_type;
		switch(c_type)
		{
		case TYPE_FDS:
			memcpy(Inst[i].fds_wave, &buf[dp], 64); dp += 64;
			memcpy(Inst[i].fds_mtable, &buf[dp], 32); dp += 32;
			memcpy(&Inst[i].fds_mspeed, &buf[dp], 4); dp += 4;
			memcpy(&Inst[i].fds_mdepth, &buf[dp], 4); dp += 4;
			memcpy(&Inst[i].fds_mdelay, &buf[dp], 4); dp += 4;
			for(j=0;j<3;j++){
				seqlen = buf[dp++];
				Inst[i].Seq[j].length = seqlen;
				memcpy(&Inst[i].Seq[j].loop, &buf[dp], 4); dp += 4;
				memcpy(&Inst[i].Seq[j].release, &buf[dp], 4); dp += 4;
				memcpy(&Inst[i].Seq[j].setting, &buf[dp], 4); dp += 4;
				memcpy(Inst[i].Seq[j].data, &buf[dp], seqlen); dp += seqlen;
			}
			break;

		case TYPE_VRC7: 
			memcpy(&Inst[i].vrc7_patch, &buf[dp], 4); dp += 4;
			memcpy(Inst[i].vrc7_data, &buf[dp], 8); dp += 8;		
			break;
			
		case TYPE_2A03:
		case TYPE_VRC6:
		case TYPE_N163:
			memcpy(&Inst[i].seq_count, &buf[dp], 4); dp += 4;
			memcpy(Inst[i].seq_data, &buf[dp], 10); dp += 10;
			if(c_type == TYPE_2A03){
				memcpy(Inst[i].DPCM_data, &buf[dp], 288);
				dp += 288;
			} else if(c_type == TYPE_N163){
				memcpy(&Inst[i].wave_size, &buf[dp], 4); dp += 4;
				memcpy(&Inst[i].wave_pos, &buf[dp], 4); dp += 4;
				memcpy(&Inst[i].wave_count, &buf[dp], 4); dp += 4;
				for(j=0;j<Inst[i].wave_count;j++){		
					memcpy(Inst[i].wave_data[j], &buf[dp], Inst[i].wave_size);
					dp += Inst[i].wave_size;				
				}
			}
			break;

		default: 
			printf("Error: Instrument %i has unknown instrument type\n", i);
			printf("offset: 0x%X\n", dp);
			return false;
		}
		
		memcpy(&namelen, &buf[dp], 4); dp += 4;
		Inst[i].name_length = namelen;
		memcpy(Inst[i].name, &buf[dp], namelen); dp += namelen;		
	}
	
	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;	
	memcpy(&n_sequences, &buf[dp], 4); dp += 4;
	
	for(i=0;i<n_sequences;i++){
		memcpy(&Sequence[i].index, &buf[dp], 4); dp += 4;
		memcpy(&Sequence[i].type, &buf[dp], 4); dp += 4;
		seqlen = buf[dp++];
		Sequence[i].length = seqlen;
		memcpy(&Sequence[i].loop, &buf[dp], 4); dp += 4;
		memcpy(Sequence[i].data, &buf[dp], seqlen); dp += seqlen;		
	}
	
	for(i=0;i<n_sequences;i++){
		memcpy(&Sequence[i].release, &buf[dp], 4); dp += 4;
		memcpy(&Sequence[i].setting, &buf[dp], 4); dp += 4;
	}

	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;
	memcpy(&n_frames, &buf[dp], 4); dp += 4;
	memcpy(&speed, &buf[dp], 4); dp += 4;
	memcpy(&tempo, &buf[dp], 4); dp += 4;
	memcpy(&p_length, &buf[dp], 4); dp += 4;
	for(i=0;i<n_frames;i++){
		memcpy(fdata[i], &buf[dp], channels);
		dp += channels;		
	}

	memcpy(block_id, &buf[dp], 16); dp += 16;
	memcpy(&block_version, &buf[dp], 4); dp += 4;
	memcpy(&block_size, &buf[dp], 4); dp += 4;	
	patterns_end = dp + block_size;
	n_patterns = 0;
	for(i=0;dp<patterns_end;i++){
		memcpy(&Pattern[i].track, &buf[dp], 4); dp += 4;
		memcpy(&Pattern[i].channel, &buf[dp], 4); dp += 4;
		memcpy(&Pattern[i].index, &buf[dp], 4); dp += 4;;
		memcpy(&Pattern[i].items, &buf[dp], 4); dp += 4;
		int item_size = 10 + 2*ch_fx[Pattern[i].channel];
		for(j=0;j<Pattern[i].items;j++){
			memcpy(&Pattern[i].Item[j], &buf[dp], item_size);
			dp += item_size;
		}
		n_patterns++;
	}
	return true;
}

bool CFTMFile::SaveFile(char* filename)
{
	int version;
	int blocksize;
	int count;
	int namelen;
	int i, j;
	
	fp_dst = fopen(filename,"wb"); if(fp_dst==NULL) {printf("Error! Source file not found. :(\n"); return false;}

	version = 0x440;	
	fwrite("FamiTracker Module", 1, 18, fp_dst);	
	fwrite(&version, 4, 1, fp_dst);
	
	version = 5;
	blocksize = (expansion&0x10)? 33: 29;
	fwrite("PARAMS\0\0\0\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&expansion, 1, 1, fp_dst);
	fwrite(&channels, 4, 1, fp_dst);
	fwrite(&machine, 4, 1, fp_dst);
	fwrite(&e_speed, 4, 1, fp_dst);
	fwrite(&v_style, 4, 1, fp_dst);
	fwrite(&highlight1, 4, 1, fp_dst);
	fwrite(&highlight2, 4, 1, fp_dst);
	if(expansion&0x10) fwrite(&N163_channels, 4, 1, fp_dst);
	fwrite(&s_split, 4, 1, fp_dst);

	version = 1;
	blocksize = 96;	
	fwrite("INFO\0\0\0\0\0\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(author, 1, 32, fp_dst);
	fwrite(title, 1, 32, fp_dst);
	fwrite(ccccc, 1, 32, fp_dst);
	
	version = 3;
	blocksize = 2 + 2*channels;
	fwrite("HEADER\0\0\0\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&n_tracks, 1, 1, fp_dst);
	//fwrite("New song", 1, 9, fp_dst);
	fwrite("\0", 1, 1, fp_dst);
	for (i=0;i<channels;i++){
		fwrite(&i, 1, 1, fp_dst);
		fwrite(&ch_fx[i], 1, 1, fp_dst);
	}


	version = 6;
	blocksize = 0x4;
	for(i=0;i<n_instruments;i++){
		switch(Inst[i].type){
		case TYPE_2A03: blocksize += 0x145; break;
		case TYPE_VRC6: blocksize += 0x25; break;
		case TYPE_VRC7: blocksize += 0x23; break;
		case TYPE_S5B: blocksize += 0x25; break;
		case TYPE_FDS:
			blocksize += 0xAA;
			for(j=0;j<3;j++) blocksize += Inst[i].Seq[j].length;
			break;
		case TYPE_N163:
			blocksize += 0x31;
			blocksize += Inst[i].wave_size * Inst[i].wave_count;
			break;
		}
	}
	
	fwrite("INSTRUMENTS\0\0\0\0\0", 1, 16,fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&n_instruments, 4, 1, fp_dst);
	
	namelen = 0xE;

	for(i=0;i<n_instruments;i++){
		fwrite(&Inst[i].index, 4, 1, fp_dst);
		fwrite(&Inst[i].type, 1, 1, fp_dst);
		int c_type = Inst[i].type;
		switch(c_type)
		{
		case TYPE_FDS:
			fwrite(Inst[i].fds_wave, 64, 1, fp_dst);
			fwrite(Inst[i].fds_mtable, 32, 1, fp_dst);
			fwrite(&Inst[i].fds_mspeed, 4, 1, fp_dst);
			fwrite(&Inst[i].fds_mdepth, 4, 1, fp_dst);
			fwrite(&Inst[i].fds_mdelay, 4, 1, fp_dst);
			for(j=0;j<3;j++){
				fwrite(&Inst[i].Seq[j].length, 1, 1, fp_dst);
				fwrite(&Inst[i].Seq[j].loop, 4, 1, fp_dst);
				fwrite(&Inst[i].Seq[j].release, 4, 1, fp_dst);
				fwrite(&Inst[i].Seq[j].setting, 4, 1, fp_dst);
				fwrite(Inst[i].Seq[j].data, Inst[i].Seq[j].length, 1, fp_dst);
			}
			break;

		case TYPE_VRC7: 
			fwrite(&Inst[i].vrc7_patch, 4, 1, fp_dst);
			fwrite(Inst[i].vrc7_data, 1, 8, fp_dst);	
			break;
			
		case TYPE_2A03:
		case TYPE_VRC6:
		case TYPE_N163:
		default:
			fwrite(&Inst[i].seq_count, 4, 1, fp_dst);
			fwrite(Inst[i].seq_data, 1, 10, fp_dst);
			if(c_type == TYPE_2A03){
				fwrite(Inst[i].DPCM_data, 1, 288, fp_dst);
			} else if(c_type == TYPE_N163){
				fwrite(&Inst[i].wave_size, 4, 1, fp_dst);
				fwrite(&Inst[i].wave_pos, 4, 1, fp_dst);
				fwrite(&Inst[i].wave_count, 4, 1, fp_dst);
				for(j=0;j<Inst[i].wave_count;j++) fwrite(Inst[i].wave_data[j], 1, Inst[i].wave_size, fp_dst);
			}
			break;
		}
		fwrite(&namelen, 4, 1, fp_dst);
		fwrite("New instrument", 1, 14, fp_dst);
	}
	
	version = 6;
	blocksize = 4;
	for(i=0;i<n_sequences;i++) blocksize += 21 + Sequence[i].length;
	fwrite("SEQUENCES\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&n_sequences, 4, 1, fp_dst);
	
	for(i=0;i<n_sequences;i++){
		fwrite(&Sequence[i].index, 4, 1, fp_dst);
		fwrite(&Sequence[i].type, 4, 1, fp_dst);
		fwrite(&Sequence[i].length, 1, 1, fp_dst);
		fwrite(&Sequence[i].loop, 4, 1, fp_dst);
		fwrite(Sequence[i].data, 1, Sequence[i].length, fp_dst);		
	}
	
	for(i=0;i<n_sequences;i++){
		fwrite(&Sequence[i].release, 4, 1, fp_dst);
		fwrite(&Sequence[i].setting, 4, 1, fp_dst);
	}

	version = 3;
	blocksize = 0x10 + channels*n_frames;
	fwrite("FRAMES\0\0\0\0\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&n_frames, 4, 1, fp_dst);
	fwrite(&speed, 4, 1, fp_dst);
	fwrite(&tempo, 4, 1, fp_dst);
	fwrite(&p_length, 4, 1, fp_dst);
	for(i=0;i<n_frames;i++)
		fwrite(fdata[i], 1, channels, fp_dst);

	version = 5;
	blocksize = 0;
	for (i=0;i<n_patterns;i++){		
		if(!Pattern[i].items) continue;
		blocksize += 16 + (10+2*ch_fx[Pattern[i].channel])*Pattern[i].items;
	}

	fwrite("PATTERNS\0\0\0\0\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	
	for(i=0;i<n_patterns;i++){
		if(!Pattern[i].items) continue;
		fwrite(&Pattern[i].track, 4, 1, fp_dst);
		fwrite(&Pattern[i].channel, 4, 1, fp_dst);
		fwrite(&Pattern[i].index, 4, 1, fp_dst);
		fwrite(&Pattern[i].items, 4, 1, fp_dst);
		int item_size = 10+2*ch_fx[Pattern[i].channel];
		for(j=0;j<Pattern[i].items;j++){
			fwrite(&Pattern[i].Item[j], 1, item_size, fp_dst);
		}
	}

	version = 1;
	blocksize = 1 + 9*n_samples;
	for(i=0;i<n_samples;i++) blocksize += 0xE + Sample[i].datalength;
	fwrite("DPCM SAMPLES\0\0\0\0", 1, 16, fp_dst);
	fwrite(&version, 4, 1, fp_dst);
	fwrite(&blocksize, 4, 1, fp_dst);
	fwrite(&n_samples, 1, 1, fp_dst);

	for (i=0;i<n_samples;i++){
		fwrite(&i, 1, 1, fp_dst);
		fwrite(&namelen, 4, 1, fp_dst);
		fwrite("New instrument", 1, 14, fp_dst);
		fwrite(&Sample[i].datalength, 4, 1, fp_dst);
		fwrite(Sample[i].data, Sample[i].datalength, 1, fp_dst);
	}
	
	if(expansion & 0x01){
		version = 6;
		blocksize = 4;
		for(i=0;i<n_VRC6_sequences;i++) blocksize += 21 + VRC6_Sequence[i].length;
		fwrite("SEQUENCES_VRC6\0\0", 1, 16, fp_dst);
		fwrite(&version, 4, 1, fp_dst);
		fwrite(&blocksize, 4, 1, fp_dst);
		fwrite(&n_VRC6_sequences, 4, 1, fp_dst);
		
		for(i=0;i<n_VRC6_sequences;i++){
			fwrite(&VRC6_Sequence[i].index, 4, 1, fp_dst);
			fwrite(&VRC6_Sequence[i].type, 4, 1, fp_dst);
			fwrite(&VRC6_Sequence[i].length, 1, 1, fp_dst);
			fwrite(&VRC6_Sequence[i].loop, 4, 1, fp_dst);
			fwrite(VRC6_Sequence[i].data, 1, VRC6_Sequence[i].length, fp_dst);		
		}
		
		for(i=0;i<n_VRC6_sequences;i++){
			fwrite(&VRC6_Sequence[i].release, 4, 1, fp_dst);
			fwrite(&VRC6_Sequence[i].setting, 4, 1, fp_dst);
		}
	}
	
	if(expansion & 0x10){
		version = 1;
		blocksize = 4;
		for(i=0;i<n_N163_sequences;i++) blocksize += 21 + N163_Sequence[i].length;
		fwrite("SEQUENCES_N163\0\0", 1, 16, fp_dst);
		fwrite(&version, 4, 1, fp_dst);
		fwrite(&blocksize, 4, 1, fp_dst);
		fwrite(&n_N163_sequences, 4, 1, fp_dst);
		
		for(i=0;i<n_N163_sequences;i++){
			fwrite(&N163_Sequence[i].index, 4, 1, fp_dst);
			fwrite(&N163_Sequence[i].type, 4, 1, fp_dst);
			fwrite(&N163_Sequence[i].length, 1, 1, fp_dst);
			fwrite(&N163_Sequence[i].loop, 4, 1, fp_dst);
			fwrite(&N163_Sequence[i].release, 4, 1, fp_dst);
			fwrite(&N163_Sequence[i].setting, 4, 1, fp_dst);
			fwrite(N163_Sequence[i].data, 1, N163_Sequence[i].length, fp_dst);		
		}	
	}
	
	
	
	
/*

bool CFTMFile::WriteVRC6Sequences()
{
	if(!(expansion&0x1)) return true;
	fwrite("SEQUENCES\0\0\0\0\0\0\0",1,16,fp_dst);
	write_dword(fp_dst,999);
	int pblocksize=4; for(int i=0;i<n_VRC6_sequences;i++){pblocksize+=21+VRC6_Sequence[i].length;}
	write_dword(fp_dst,pblocksize);
	write_dword(fp_dst,n_VRC6_sequences);
	for(int i=0;i<n_VRC6_sequences;i++){
		write_dword(fp_dst,VRC6_Sequence[i].index);
		write_dword(fp_dst,VRC6_Sequence[i].type);
		write_byte(fp_dst,VRC6_Sequence[i].length);
		write_dword(fp_dst,VRC6_Sequence[i].loop);
		fwrite(VRC6_Sequence[i].data,1,VRC6_Sequence[i].length,fp_dst);		
	}

	for(int i=0;i<n_VRC6_sequences;i++){
		write_dword(fp_dst,VRC6_Sequence[i].release);
		write_dword(fp_dst,VRC6_Sequence[i].setting);
	}

	return true;
}

bool CFTMFile::WriteN163Sequences()
{
	if(!(expansion&0x10)) return true;
	if(!n_N163_sequences) return true;
	fwrite("SEQUENCES\0\0\0\0\0\0\0",1,16,fp_dst);
	write_dword(fp_dst,999);
	int pblocksize=4; for(int i=0;i<n_N163_sequences;i++){pblocksize+=21+N163_Sequence[i].length;}
	write_dword(fp_dst,pblocksize);
	write_dword(fp_dst,n_N163_sequences);
	for(int i=0;i<n_N163_sequences;i++){
		write_dword(fp_dst,N163_Sequence[i].index);
		write_dword(fp_dst,N163_Sequence[i].type);
		write_byte(fp_dst,N163_Sequence[i].length);
		write_dword(fp_dst,N163_Sequence[i].loop);
		write_dword(fp_dst,N163_Sequence[i].release);
		write_dword(fp_dst,N163_Sequence[i].setting);
		fwrite(N163_Sequence[i].data,1,N163_Sequence[i].length,fp_dst);		
	}
	return true;
}*/

	fwrite("\0END", 1, 4, fp_dst);
	fclose(fp_dst);
	return true;
}

bool CFTMFile::ExportTXT(char* filename)
{
	int i,j,k;
	FTItem Cell;

	fp_dst=fopen(filename,"w"); if(fp_dst==NULL) {printf("Error! Could not create log file. :(\n"); return false;}
	fprintf(fp_dst, "Expansion: 0x%X\n", expansion);
	fprintf(fp_dst, "Channels: %i\n", channels);
	fprintf(fp_dst, "N163 channels: %i\n", N163_channels);
	fprintf(fp_dst, "Instruments: %i\n", n_instruments);
	fprintf(fp_dst, "2A03 sequences: %i\n", n_sequences);
	fprintf(fp_dst, "Frames: %i\n", n_frames);
	fprintf(fp_dst, "Patterns: %i\n", n_patterns);
	fprintf(fp_dst, "Samples: %i\n", n_samples);

	fprintf(fp_dst,"\nINSTRUMENTS:\n\n");

	for(i=0;i<n_instruments;i++){
		fprintf(fp_dst,"Instrument %i index: %i\n",i,Inst[i].index);
		fprintf(fp_dst,"Instrument %i type: %i\n",i,Inst[i].type);
		fprintf(fp_dst,"Instrument %i sequence count: %i\n",i,Inst[i].seq_count);
		fprintf(fp_dst,"Instrument %i sequence data:",i);
		for(j=0;j<2*SEQ_COUNT;j++){
			fprintf(fp_dst," %02X",Inst[i].seq_data[j]);
		}
		fprintf(fp_dst,"\n");
		fprintf(fp_dst,"Instrument %i name length: %i\n",i,Inst[i].name_length);
	}


	fprintf(fp_dst,"\nSEQUENCES:\n\n");

	for(i=0;i<n_sequences;i++){
		fprintf(fp_dst,"Sequence %i index: %i\n",i,Sequence[i].index);
		fprintf(fp_dst,"Sequence %i type: %i\n",i,Sequence[i].type);
		fprintf(fp_dst,"Sequence %i length: %i\n",i,Sequence[i].length);
	}

	if(expansion&0x10){
		fprintf(fp_dst,"\nN163 SEQUENCES:\n\n");

		for(i=0;i<n_sequences;i++){
			fprintf(fp_dst,"N163 Sequence %i index: %i\n",i,N163_Sequence[i].index);
			fprintf(fp_dst,"N163 Sequence %i type: %i\n",i,N163_Sequence[i].type);
			fprintf(fp_dst,"N163 Sequence %i length: %i\n",i,N163_Sequence[i].length);
		}
	}

	fprintf(fp_dst,"\nFRAMES:\n\n");

	for(i=0;i<n_frames;i++){
		fprintf(fp_dst,"%02X |",i);
		for(j=0;j<channels;j++){
			fprintf(fp_dst," %02X",fdata[i][j]);
		}
		fprintf(fp_dst,"\n");
	}

	fprintf(fp_dst,"\nPATTERNS:\n\n");

	for(i=0;i<n_patterns;i++){
		fprintf(fp_dst,"Pattern %i track: %i\n",i,Pattern[i].track);
		fprintf(fp_dst,"Pattern %i channel: %i\n",i,Pattern[i].channel);
		fprintf(fp_dst,"Pattern %i index: %i\n",i,Pattern[i].index);
		fprintf(fp_dst,"Pattern %i items: %i\n",i,Pattern[i].items);
	}

	fprintf(fp_dst,"\nSAMPLES:\n\n");

	for(i=0;i<n_samples;i++){
		fprintf(fp_dst,"Sample %i index: %i\n",i,Sample[i].index);
		fprintf(fp_dst,"Sample %i name length: %i\n",i,Sample[i].namelength);
		fprintf(fp_dst,"Sample %i data length: %i\n",i,Sample[i].datalength);
	}

	fclose(fp_dst);

	//return true;

	fp_dst=fopen("patterns.log","w"); if(fp_dst==NULL) {printf("Error! Could not create log file. :(\n"); return false;}

	for(i=0;i<n_patterns;i++){
		fprintf(fp_dst,"Pattern %i track: %i\n",i,Pattern[i].track);
		fprintf(fp_dst,"Pattern %i channel: %i\n",i,Pattern[i].channel);
		fprintf(fp_dst,"Pattern %i index: %i\n",i,Pattern[i].index);
		fprintf(fp_dst,"Pattern %i items: %i\n\n",i,Pattern[i].items);
		for(j=0;j<Pattern[i].items;j++){
			memcpy(&Cell,&Pattern[i].Item[j],sizeof(FTItem));
			fprintf(fp_dst,"%02X | %c%c%c %02X %01X %c%02X %c%02X %c%02X %c%02X\n",Cell.row,ftmn2not(Cell.note),ftmn2acc(Cell.note),ftmno2oct(Cell.note,Cell.octave),Cell.instrument,Cell.volume&0xF,ftmfx(Cell.fxdata[0]),Cell.fxdata[1],ftmfx(Cell.fxdata[2]),Cell.fxdata[3],ftmfx(Cell.fxdata[4]),Cell.fxdata[5],ftmfx(Cell.fxdata[6]),Cell.fxdata[7]);
		}
		fprintf(fp_dst,"\n");
	}

	fclose(fp_dst);
	return true;
}

bool CFTMFile::Optimize()
{
	FTPattern Temp;
	FTItem Prev;

	for(int i=0;i<n_patterns;i++){
		Temp.track=Pattern[i].track;
		Temp.channel=Pattern[i].channel;
		Temp.index=Pattern[i].index;
		Temp.items=0;
		
		memset(&Prev, 0, sizeof(FTItem));
		Prev.instrument = 64;
		Prev.volume = 16;
		
		memcpy(&Temp.Item[Temp.items], &Prev, sizeof(FTItem));

		for(int j=0;j<Pattern[i].items;j++){
			bool non_empty=false;
			bool new_item=true;

			if(Pattern[i].Item[j].note!=0) non_empty=true;
			if(Pattern[i].Item[j].instrument!=64) non_empty=true;
			if(Pattern[i].Item[j].volume!=16) non_empty=true;
			for(int k=0;k<8;k+=2){
				if(Pattern[i].Item[j].fxdata[k]!=0) non_empty=true;
			}

			if(!non_empty) continue;

			if(Pattern[i].Item[j].note!=Prev.note) new_item=true;
			if(Pattern[i].Item[j].octave!=Prev.octave) new_item=true;
			if(Pattern[i].Item[j].instrument!=Prev.instrument) new_item=true;
			if(Pattern[i].Item[j].volume!=Prev.volume) new_item=true;
			for(int k=0;k<8;k++){
				if(Pattern[i].Item[j].fxdata[k]!=Prev.fxdata[k]) new_item=true;
			}

			if(!new_item) continue;
			
			memcpy(&Prev, &Pattern[i].Item[j], sizeof(FTItem));
			memcpy(&Temp.Item[Temp.items], &Prev, sizeof(FTItem));
			Temp.items++;
		}
		memcpy(&Pattern[i], &Temp, sizeof(FTPattern));
	}
	return true;
}


bool CFTMFile::LogStuff()
{
	fprintf(fp_log,"\nFRAMES\n\n");

	for(int i=0;i<n_frames;i++){
		fprintf(fp_log,"%02X:",i);
		for(int j=0;j<channels;j++) fprintf(fp_log," %02X",fdata[i][j]);
		fprintf(fp_log,"\n");
	}

	fprintf(fp_log,"\nPATTERNS\n\n");

	for(int i=0;i<n_patterns;i++){
		fprintf(fp_log,"Pattern %i index: %i channel %i: items: %i\n",i,Pattern[i].index,Pattern[i].channel,Pattern[i].items);
	}

	return true;
}

int CFTMFile::ScanActualLength()
{
	FILE* fp_len=fopen("length.log","w");

	byte visited[MAX_FRAMES];
	int jumping = -1;
	int skipping = 0;
	int c_speed = speed;
	int c_tempo = tempo;
	int c_pattern = 0;
	int spr = (MAGIC_NUMBER*c_speed)/c_tempo; //Samples per row
	int total_samples = 0;

	memset(visited, 0, MAX_FRAMES);

	for(int i=0;i<n_frames;i++){
		if(visited[i]) break;
		visited[i] = 1;
		fprintf(fp_len,"Visiting frame: %i speed: %i tempo: %i spr: %i samples: %i\n",i,c_speed,c_tempo,spr,total_samples);
		for (int k=0;k<p_length;k++){
			for (int j=0; j<channels; j++){
				int c_pattern = FindPattern(j,fdata[i][j]);
				if(c_pattern==-1) {printf("Oopsie! %i %i %i :(\n",i,j,fdata[i][j]); return -1;}

				for (int l=0;l<ch_fx[j]+1;l++) {
					switch (Pattern[c_pattern].Item[k].fxdata[2*l]){
					case EF_SPEED: 
						if(Pattern[c_pattern].Item[k].fxdata[1+2*l]>=32)
							c_tempo = Pattern[c_pattern].Item[k].fxdata[1+2*l];
						else
							c_speed = Pattern[c_pattern].Item[k].fxdata[1+2*l];
						if(c_speed<1) c_speed = 1;
						spr = (MAGIC_NUMBER*c_speed)/c_tempo;
						break;
					case EF_JUMP: jumping = Pattern[c_pattern].Item[k].fxdata[1+2*l]; break;
					case EF_SKIP: skipping = 1; break;
					case EF_HALT: total_samples += spr; goto length_found;
					}
				}
			}
			total_samples += spr;
			if(jumping != -1){
				i = jumping-1;
				jumping = -1;
				skipping = 0;
				break;
			}
			if(skipping){
				skipping = 0;
				break;
			}
		}
	}
	
length_found:

	fclose(fp_len);
	return total_samples;
}

int CFTMFile::FindPattern(int ch, int index)
{
	int c_pattern = -1;

	for(int i=0;i<n_patterns;i++){
		if(Pattern[i].index != index) continue;
		if(Pattern[i].channel != ch) continue;
		c_pattern = i;
		break;
	}

	return c_pattern;
}