#pragma once

#include <stdio.h>
#include "global_stuff.h"

#define SEQ_COUNT 5
#define MAX_FRAMES 128
#define VRC7_REGS 8

struct stSequence{	
	int index;
	int type;
	unsigned char length;
	int loop;
	int release;
	int setting;
	unsigned char data[256];
	stSequence();
};

struct Instrument {
	int index;
	unsigned char type;
	int seq_count;
	unsigned char seq_data[2*SEQ_COUNT];
	unsigned char DPCM_data[288];
	int name_length;
	char name[64];
	unsigned char fds_wave[64];
	unsigned char fds_mtable[32];
	int fds_mspeed;
	int fds_mdepth;
	int fds_mdelay;
	stSequence Seq[3];
	int vrc7_patch;
	unsigned char vrc7_data[8];
	int wave_size;
	int wave_pos;
	int wave_count;
	unsigned char wave_data[16][32];
	Instrument();
};

struct DPCM_Sample{
	int index;
	int namelength;
	int datalength;
	char name[64];
	unsigned char data[4096];
};

struct FTItem{
	int row;
	unsigned char note;
	unsigned char octave;
	unsigned char instrument;
	unsigned char volume;
	unsigned char fxdata[8];
	FTItem();
};

struct FTPattern{
	int track;
	int channel;
	int index;
	int   items;
	FTItem Item[256];
	FTPattern();
};

class CFTMFile{
public:	CFTMFile();

	FILE* fp_dst;
	FILE* fp_log;

	int src_size;

	char file_id[24];
	int file_version;

	char pr_id[24]; //PARAMS
	int pr_version;
	int pr_size;
	unsigned char expansion;
	int channels;
	int machine;
	int e_speed;
	int v_style;
	int highlight1;
	int highlight2;
	int N163_channels;
	int s_split;

	char author[32];
	char title[32];
	char ccccc[32];

	unsigned char n_tracks;
	char track_name[64];
	char ch_id[32];
	char ch_fx[32];

	int n_instruments;
	Instrument Inst[64];

	int n_sequences;
	stSequence Sequence[320];

	int n_frames;
	int speed;
	int tempo;
	int p_length;
	unsigned char fdata[MAX_FRAMES][32];

	FTPattern Pattern[960];

	unsigned char n_samples;

	int	n_VRC6_sequences;
	stSequence VRC6_Sequence[320];

	int	n_N163_sequences;
	stSequence N163_Sequence[320];

	DPCM_Sample Sample[64];

	int n_patterns;

	void Reset();

	bool LoadFile(char* filename);
	bool SaveFile(char* filename);
	bool Optimize();
	bool ExportTXT(char* filename);

	bool LogStuff();

	int ScanActualLength();
	int FindPattern(int ch, int index);
};