#include <cstdio>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include "ft_stuff.h"
#include "NSF_Emu.h"
#include "NSFImport.h"

#define EXPANSION_MASK (EX_VRC6 | EX_MMC5 | EX_N163 | EX_FDS | EX_VRC7 | EX_S5B);

NSFImporter::NSFImporter()
{	
	double clock_ntsc = 1789773 / 16.0;
	double clock_pal = 1662607 / 16.0;
	double BASE_FREQ = 32.7032;
	double Freq;  // freq in Hz
	double Pitch; // pitch tegister
	int i;
	
	memset(this, 0, sizeof(NSFImporter));
	
	for(i=0;i<32;i++){
		p_ftvol[i] = 16;
		p_ftinst[i] = 64;	
	}
	
	// Init frequency tables

	for (int i=0;i<NOTE_COUNT;i++){
		Freq = BASE_FREQ * pow(2.0,i/12.0); 

		Pitch = (clock_ntsc/Freq)+0.5;
		freqtable[i] = (int)Pitch;

		Pitch = (clock_pal / Freq) - 0.5;
		paltable[i] = (int)Pitch;

		Pitch = ((clock_ntsc * 16.0) / (Freq * 14.0)) - 0.5;
		sawtable[i] = (int)Pitch;

		Pitch = (Freq * 65536.0) / (clock_ntsc / 1.0) + 0.5;
		fdstable[i] = (int)Pitch;

		Pitch = (Freq * 8 * 983040.0) / clock_ntsc;
		n163table[i] = (int)(Pitch) / 4;
		if(n163table[i]>0xFFFF) n163table[i]=0xFFFF;
	}	
}

bool NSFImporter::Import(char* nsfname, int track, int seconds)
{
	char ftmname[128];
	pNSF = new CNSF_Emu();
	pFTM = new CFTMFile();
	
	strcpy(ftmname, nsfname);
	ftmname[strlen(ftmname)-4] = 0;
	strcat(ftmname, ".ftm");

	if(!pNSF->LoadFile(nsfname)) return false;
	if(!NSF2FTM(track, seconds*60)) return false;
	if(!pFTM->Optimize()) return false;
	if(!pFTM->SaveFile(ftmname)) return false;
	return true;
}

bool NSFImporter::NSF2FTM(int track, int ticks)
{
	int c_order=0;
	int c_pattern=0;
	int c_row=0;
	int i;
	
	n_channels = 5;

	pFTM->ch_fx[0]=2;
	pFTM->ch_fx[1]=2;

	pFTM->expansion = pNSF->Info.expansion & EXPANSION_MASK;
	pFTM->N163_channels = 8;
	pFTM->n_frames = 1;
	pFTM->speed = 1;
	pFTM->tempo = 150;
	pFTM->p_length = 256;
	pFTM->n_instruments = 1;
	pFTM->Inst[0].type = 1;
	
	if(pFTM->expansion & EX_VRC6){
		vrc6_inst = pFTM->n_instruments;
		vrc6_ch = n_channels;
		pFTM->Inst[vrc6_inst].index = vrc6_inst;
		pFTM->Inst[vrc6_inst].type = TYPE_VRC6;
		pFTM->Inst[vrc6_inst].seq_count = 5;
		pFTM->n_instruments++;
		pFTM->ch_fx[vrc6_ch] = 1;
		pFTM->ch_fx[vrc6_ch+1] = 1;
		pFTM->ch_fx[vrc6_ch+2] = 1;
		n_channels += 3;
	}

	if(pFTM->expansion & EX_MMC5){
		mmc5_ch = n_channels;
		pFTM->ch_fx[n_channels]=1;
		pFTM->ch_fx[n_channels+1]=1;
		n_channels += 2;
	}

	if(pFTM->expansion & EX_N163){
		n163_ch = n_channels;
		for(i=0;i<8;i++) pFTM->ch_fx[n163_ch+i] = 1;
		n_channels += 8;
	}
	
	if(pFTM->expansion & EX_FDS){
		fds_ch = n_channels;
		//pFTM->ch_fx[fds_ch] = 0;
		n_channels += 1;
	}
	
	if(pFTM->expansion & EX_VRC7){
		vrc7_ch = n_channels;
		for(i=0;i<6;i++) pFTM->ch_fx[vrc7_ch+i] = 1;
		n_channels += 6;
	}

	if(pFTM->expansion & EX_S5B){
		s5b_inst = pFTM->n_instruments;
		s5b_ch = n_channels;
		pFTM->Inst[s5b_inst].index = s5b_inst;
		pFTM->Inst[s5b_inst].type = TYPE_S5B;
		pFTM->Inst[s5b_inst].seq_count = 5;
		pFTM->n_instruments++;
		//pFTM->ch_fx[s5b_ch] = 0;
		n_channels += 3;
	}
	
	pFTM->channels = n_channels;
	pFTM->n_patterns = n_channels * pFTM->n_frames;	

	if(!pNSF->Init(track)) return false;

	for(i=0;i<ticks;i++){
		bool w_new = true;
		bool result = pNSF->PlayOnce();
		if(!result) return false;
		c_pattern = (c_order * n_channels);
		Read_2A03(c_pattern, c_row);
		c_pattern += 5;
		
		if(pFTM->expansion & EX_VRC6){
			Read_VRC6(c_pattern, c_row);
			c_pattern += 3;
		}

		if(pFTM->expansion & EX_MMC5){
			Read_MMC5(c_pattern, c_row);
			c_pattern += 2;
		}	
		
		if(pFTM->expansion & EX_N163){
			Read_N163(c_pattern, c_row);
			c_pattern += 8;
		}				
		
		if(pFTM->expansion & EX_FDS){
			Read_FDS(c_pattern, c_row);
			c_pattern += 1;
		}
		
		if(pFTM->expansion & EX_VRC7){
			Read_VRC7(c_pattern, c_row);
			c_pattern += 6;
		}
		
		if(pFTM->expansion & EX_S5B){
			Read_S5B(c_pattern, c_row);
			c_pattern += 3;
		}
		
		c_row++;

		while(c_row>=256){			
			for(int i=0;i<n_channels;i++){
				c_pattern = (c_order*n_channels)+i;
				pFTM->Pattern[c_pattern].items   = 256;
				pFTM->Pattern[c_pattern].track   = 0;
				pFTM->Pattern[c_pattern].channel = i;
				pFTM->Pattern[c_pattern].index   = c_order;
			}
			//n_items=0;
			c_row-=256;
			c_order++;
			pFTM->n_frames++;
			pFTM->n_patterns += n_channels;
		}		
	}
	
	for(int i=0;i<n_channels;i++){
		c_pattern = (c_order*n_channels)+i;
		pFTM->Pattern[c_pattern].items   = c_row;
		pFTM->Pattern[c_pattern].track   = 0;
		pFTM->Pattern[c_pattern].channel = i;
		pFTM->Pattern[c_pattern].index   = c_order;
	}

	for(int i=0;i<pFTM->n_frames;i++){
		memset(pFTM->fdata[i],i,n_channels);
	}

	pFTM->n_samples=d_count;
	pFTM->Inst[0].type = 1;
	pFTM->Inst[0].seq_count = 5;

	for(int i=0;i<d_count;i++){
		pFTM->Inst[0].DPCM_data[3*i]=i+1;
		pFTM->Inst[0].DPCM_data[3*i+1]=0xF;
		pFTM->Inst[0].DPCM_data[3*i+2]=0xFF;
		pFTM->Sample[i].index=i;
		pFTM->Sample[i].namelength=0xE;
		pFTM->Sample[i].datalength=d_length[i];
		memcpy(pFTM->Sample[i].name,"New instrument",0xE);
		
		printf("Sample %i bank: %i address: %03X offset: %04X\n",i,d_bank[i],d_addr[i],d_offset[i]);

		if(!(pNSF->is_switched))
			//memcpy(pFTM->Sample[i].data,&pNSF->bdata[d_offset[i]],d_length[i]);
		//else
			memcpy(pFTM->Sample[i].data,&pNSF->data[d_offset[i]],d_length[i]);
	}
	
	for(int i=0;i<n163_count;i++){
		int ci = n163_inst[i];
		int cw = n163_wave[i];
		for(int j=0;j<n163_length[i]/2;j++){
			pFTM->Inst[ci].wave_data[cw][2*j] = n163_data[i][j]&15;
			pFTM->Inst[ci].wave_data[cw][2*j+1] = n163_data[i][j]>>4;
		}
	}	
	
	for(i=0;i<n163_instruments;i++){
		int ci = n163_inst_idx[i];
		printf("N163 instrument %i index: %i\n", ci, pFTM->Inst[ci].index);
		printf("N163 instrument %i wave size: %i\n", ci, pFTM->Inst[ci].wave_size);
		printf("N163 instrument %i wave pos: %i\n", ci, pFTM->Inst[ci].wave_pos);
		printf("N163 instrument %i wave count: %i\n", ci, pFTM->Inst[ci].wave_count);
		printf("N163 instrument %i type: %i\n", ci, pFTM->Inst[ci].type);
		printf("N163 instrument %i seq count: %i\n", ci, pFTM->Inst[ci].seq_count);	
	}

	for(i=0;i<fds_count;i++){
		int ci = fds_inst_idx[i];
		printf("FDS instrument %i index: %i\n", ci, pFTM->Inst[ci].index);
		printf("FDS instrument %i type: %i\n", ci, pFTM->Inst[ci].type);
	}
	
	return true;
}

void NSFImporter::Read_2A03(int first_pattern, int c_row)
{
	FTItem* Item;
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int nes_sweep=0;
	int dpcm_status=0;
	int dpcm_addr=0;
	int dpcm_length=0;
	int dpcm_earrape=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;
	int c_offset=0;
	int c_sample = 0;
	int c_pattern = first_pattern;
	int i;

	for(i=0;i<2;i++){
		c_pattern = first_pattern + i;
		nes_vol   = pNSF->data[0x4000 + 4*i]&0xF;
		nes_duty  = pNSF->data[0x4000 + 4*i]>>6;
		nes_pitch = ((pNSF->data[0x4003 + 4*i]&7)<<8) | pNSF->data[0x4002 + 4*i];
		GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);
		
		Item = &pFTM->Pattern[c_pattern].Item[c_row];

		Item->row		= c_row;
		Item->note       = c_ftnote;
		Item->octave     = c_ftoct;
		Item->instrument = 0;
		Item->volume     = nes_vol;
		
		if(c_ftpitch != p_ftfx1[i]){
			Item->fxdata[0] = EF_PITCH;
			Item->fxdata[1] = c_ftpitch;
		}
		if(nes_duty != p_ftfx2[i]){
			Item->fxdata[2] = EF_DUTY_CYCLE;
			Item->fxdata[3] = nes_duty;
		}
		//pFTM->Pattern[c_pattern].Item[c_row].fxdata[4]  = (gb_sweep==p_ftsweep)? 0: (gb_sweep&0x8)? EF_SWEEPUP: EF_SWEEPDOWN;
		//pFTM->Pattern[c_pattern].Item[c_row].fxdata[5]  = (gb_sweep==p_ftsweep && (gb_sweep&0x77))? 0: nes_sweep&0x77;
		
		if(p_ftnote[i] == c_ftnote && p_ftoct[i] == c_ftoct){
			Item->note = 0;
			Item->octave = 0;
			Item->instrument = 64;
		}
		
		if(p_ftvol[i] == nes_vol) Item->volume = 16;

		p_ftnote[i] = c_ftnote;
		p_ftoct[i] = c_ftoct;
		p_ftvol[i] = nes_vol;
		p_ftfx1[i] = c_ftpitch;
		p_ftfx2[i] = nes_duty;				
	}

	c_pattern = first_pattern+2;
	Item = &pFTM->Pattern[c_pattern].Item[c_row];

	nes_vol   = (pNSF->data[0x4008]&0x7F) | (pNSF->data[0x400B]>>3)? 15: 0;
	nes_pitch = ((pNSF->data[0x400B]&7)<<8) | pNSF->data[0x400A];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	Item->row		= c_row;
	Item->note       = c_ftnote;
	Item->octave     = c_ftoct;
	Item->instrument = 0;
	Item->volume     = nes_vol? 15: 0;
	
	if(c_ftpitch != p_ftfx1[2]){
		Item->fxdata[0] = EF_PITCH;
		Item->fxdata[1] = c_ftpitch;
	}
	
	if(p_ftnote[2] == c_ftnote && p_ftoct[2] == c_ftoct){
		Item->note = 0;
		Item->octave = 0;
		Item->instrument = 64;
	}
	
	if(p_ftvol[2] == nes_vol) Item->volume = 16;

	p_ftnote[2] = c_ftnote;
	p_ftoct[2] = c_ftoct;
	p_ftvol[2] = nes_vol;
	p_ftfx1[2] = c_ftpitch;

	c_pattern = first_pattern+3;
	Item = &pFTM->Pattern[c_pattern].Item[c_row];

	nes_vol   = pNSF->data[0x400C]&0xF;
	nes_pitch = 31-(pNSF->data[0x400E]&0xF);
	nes_duty = pNSF->data[0x400E]>>7;

	c_ftnote = (nes_pitch%12)+1;
	c_ftoct = nes_pitch/12;

	Item->row = c_row;
	Item->note = c_ftnote;
	Item->octave = c_ftoct;
	Item->instrument = 0;
	Item->volume = nes_vol;
	if(nes_duty != p_ftfx1[3]){
		Item->fxdata[0] = EF_DUTY_CYCLE;
		Item->fxdata[1] = nes_duty;
	}

	if(p_ftnote[3] == c_ftnote && p_ftoct[3] == c_ftoct){
		Item->note = 0;
		Item->octave = 0;
		Item->instrument = 64;
	}
	
	if(p_ftvol[3] == nes_vol) Item->volume = 16;
	
	p_ftnote[3] = c_ftnote;
	p_ftoct[3] = c_ftoct;
	p_ftvol[3] = nes_vol;
	p_ftfx1[3] = nes_duty;

	// DPCM

	c_pattern = first_pattern+4;
	Item = &pFTM->Pattern[c_pattern].Item[c_row];

	dpcm_status = pNSF->dpcm_triggered;
	pNSF->dpcm_triggered &= 0;

	if(dpcm_status){
		bool new_sample=true;
		int bank = 4+(pNSF->data[0x4012]>>6);
		int c_addr = 64*(pNSF->data[0x4012]&0x3F);
		for(int i=0;i<d_count;i++){
			if(pNSF->is_switched && pNSF->is_FDS){
				c_offset = 4096*pNSF->c_bank[bank]+c_addr;
			} else if(pNSF->is_switched && !pNSF->is_FDS){
				c_offset = 4096*pNSF->c_bank[bank]+c_addr;			
			} else {
				c_offset = 0xC000+64*pNSF->data[0x4012];				
			}
			if(c_offset != d_offset[i]) continue;
			c_sample=i;
			new_sample=false;
			break;
		}

		if(new_sample){
			if(pNSF->is_switched){
				d_bank[d_count] = pNSF->c_bank[4+(pNSF->data[0x4012]>>6)];
				d_addr[d_count] = 64 * (pNSF->data[0x4012]&0x3F);
				d_offset[d_count] = 4096*d_bank[d_count] + d_addr[d_count];
			} else {
				d_bank[d_count] = 4+(pNSF->data[0x4012]>>6);
				d_addr[d_count] = 64*(pNSF->data[0x4012]&0x3F);
				d_offset[d_count] = 0xC000 + 64*pNSF->data[0x4012];				
			}			
			d_length[d_count] = 16*pNSF->data[0x4013]+1;
			if(pNSF->is_switched) memcpy(pFTM->Sample[d_count].data,&pNSF->data[0xC000+64*pNSF->data[0x4012]],d_length[d_count]);
			c_sample = d_count++;
		}
	}

	c_ftpitch = pNSF->data[0x4010]&0xF;
	c_ftnote = dpcm_status? ((c_sample)%12)+1: 0;
	c_ftoct  = dpcm_status? (c_sample)/12: 0;

	Item->row = c_row;
	Item->instrument = 64;
	
	if(dpcm_status){
		Item->note = c_ftnote;
		Item->octave = c_ftoct;
		Item->instrument = 0;
	}
	
	if(c_ftpitch != p_ftfx1[4]){
		Item->fxdata[0] = EF_DPCM_PITCH;
		Item->fxdata[1] = c_ftpitch;
	}

	p_ftfx1[4] = c_ftpitch;
}

void NSFImporter::Read_VRC6(int first_pattern, int c_row)
{
	FTItem* Item;
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;
	int c_pattern;
	int ch;
	int i;
	
	for(i=0;i<2;i++){
		ch = vrc6_ch + i;
		c_pattern = first_pattern + i;
		Item = &pFTM->Pattern[c_pattern].Item[c_row];
		
		nes_vol   = pNSF->vrc6_vol[i]&0xF;
		nes_duty  = (pNSF->vrc6_vol[i]&0x70)>>4;
		nes_pitch = pNSF->vrc6_freq[i]&0xFFF;
		GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

		if(!(pNSF->vrc6_freq[i] & 0x8000)){
			c_ftnote = FT_NOTE_CUT;
			c_ftoct = 0;
			c_ftpitch = p_ftfx2[ch];
		}		

		Item->row		= c_row;
		Item->note       = c_ftnote;
		Item->octave     = c_ftoct;
		Item->instrument = vrc6_inst;
		Item->volume     = nes_vol;
		
		if(c_ftpitch != p_ftfx1[ch]){
			Item->fxdata[0] = EF_PITCH;
			Item->fxdata[1] = c_ftpitch;
		}
		if(nes_duty != p_ftfx2[ch]){
			Item->fxdata[2] = EF_DUTY_CYCLE;
			Item->fxdata[3] = nes_duty;
		}
		
		if(p_ftnote[ch] == c_ftnote && p_ftoct[ch] == c_ftoct){
			Item->note = 0;
			Item->octave = 0;
			Item->instrument = 64;
		}
		
		if(p_ftvol[ch] == nes_vol) Item->volume = 16;

		p_ftnote[ch] = c_ftnote;
		p_ftoct[ch] = c_ftoct;
		p_ftvol[ch] = nes_vol;
		p_ftfx1[ch] = c_ftpitch;
		p_ftfx2[ch] = nes_duty;				
	}

	c_pattern = first_pattern+2;
	ch = vrc6_ch + 2;
	Item = &pFTM->Pattern[c_pattern].Item[c_row];

	nes_vol   = (pNSF->vrc6_vol[2]&0x1F)>>1;
	nes_duty  = (pNSF->vrc6_vol[2]&0x3F)>>5;
	nes_pitch = pNSF->vrc6_freq[2]&0xFFF;
	GetSawNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	if(!(pNSF->vrc6_freq[2]&0x8000)){
		c_ftnote = FT_NOTE_CUT;
		c_ftoct = 0;
		c_ftpitch = p_ftfx1[ch];
	}

	Item->row		= c_row;
	Item->note       = c_ftnote;
	Item->octave     = c_ftoct;
	Item->instrument = vrc6_inst;
	Item->volume     = nes_vol;
	
	if(c_ftpitch != p_ftfx1[ch]){
		Item->fxdata[0] = EF_PITCH;
		Item->fxdata[1] = c_ftpitch;
	}
	if(nes_duty != p_ftfx2[ch]){
		Item->fxdata[2] = EF_DUTY_CYCLE;
		Item->fxdata[3] = nes_duty;
	}
	
	if(p_ftnote[ch] == c_ftnote && p_ftoct[ch] == c_ftoct){
		Item->note = 0;
		Item->octave = 0;
		Item->instrument = 64;
	}
	
	if(p_ftvol[ch] == nes_vol) Item->volume = 16;

	p_ftnote[ch] = c_ftnote;
	p_ftoct[ch] = c_ftoct;
	p_ftvol[ch] = nes_vol;
	p_ftfx1[ch] = c_ftpitch;
	p_ftfx2[ch] = nes_duty;
}

void NSFImporter::Read_MMC5(int first_pattern, int c_row)
{
	FTItem* Item;
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;
	int c_pattern;
	int ch;
	int i;
	
	for(i=0;i<2;i++){
		ch = mmc5_ch + i;
		c_pattern = first_pattern + i;
		Item = &pFTM->Pattern[c_pattern].Item[c_row];
		
		nes_vol   = pNSF->data[0x5000+4*i]&0x0F;
		nes_duty  = pNSF->data[0x5000+4*i]>>6;
		nes_pitch = ((pNSF->data[0x5003+4*i]&7)<<8) | pNSF->data[0x5002+4*i];
		GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

		/*if(!(pNSF->vrc6_freq[i] & 0x8000)){
			c_ftnote = FT_NOTE_CUT;
			c_ftoct = 0;
			c_ftpitch = p_ftfx2[ch];
		}*/

		Item->row		= c_row;
		Item->note       = c_ftnote;
		Item->octave     = c_ftoct;
		Item->instrument = vrc6_inst;
		Item->volume     = nes_vol;
		
		if(c_ftpitch != p_ftfx1[ch]){
			Item->fxdata[0] = EF_PITCH;
			Item->fxdata[1] = c_ftpitch;
		}
		if(nes_duty != p_ftfx2[ch]){
			Item->fxdata[2] = EF_DUTY_CYCLE;
			Item->fxdata[3] = nes_duty;
		}
		
		if(p_ftnote[ch] == c_ftnote && p_ftoct[ch] == c_ftoct){
			Item->note = 0;
			Item->octave = 0;
			Item->instrument = 64;
		}
		
		if(p_ftvol[ch] == nes_vol) Item->volume = 16;

		p_ftnote[ch] = c_ftnote;
		p_ftoct[ch] = c_ftoct;
		p_ftvol[ch] = nes_vol;
		p_ftfx1[ch] = c_ftpitch;
		p_ftfx2[ch] = nes_duty;				
	}
}

void NSFImporter::Read_N163(int first_pattern, int c_row)
{
	FTItem* Item;
	float Freq;
	int Register;
	int Channels = pNSF->n163_ch+1;
	int i, j, k;
	int nes_vol;
	int nes_duty;
	int nes_pitch;
	int c_ftnote;
	int c_ftoct;
	int c_ftinst;
	int c_ftvol;
	int c_ftpitch;
	int c_ftwave;
	int c_pattern;
	int c_n163inst;
	int c_wlength;
	int c_waddr;
	int ch;
	bool same_wave;
	bool new_wave;
	bool new_inst;	
	bool match;
	
	//printf("Read!\n");

	for(i=0;i<8;i++){
		//printf("i: %i\n", i);
		//printf("n163_count: %i\n", n163_count);	
	
		nes_vol = 0;
		nes_duty = 0;
		nes_pitch = 0;
		c_ftnote = 0;
		c_ftoct = 0;
		c_ftinst = 64;
		c_ftvol = 16;
		c_ftpitch = 128;
		c_ftwave = 0;
		c_pattern = first_pattern+i;

		c_ftinst = 64;
		
		//printf("Pass 1a\n");

		same_wave = true;
		new_wave = true;
		new_inst = true;
		c_n163inst = 64;
		
		//printf("i: %i\n", i);
		//printf("Pass 1aa\n");
		//printf("pNSF: 0x%08X\n", pNSF);
		
		c_wlength = 4*(64 - (pNSF->N163[i].w_length));
		c_waddr = (pNSF->N163[i].w_addr >> 1);
		ch = n163_ch + i;
		
		//printf("i: %i\n", i);
		//printf("Pass 1ab\n");

		if(n163_paddr[i] != c_waddr) same_wave=false;
		if(n163_plength[i] != c_wlength) same_wave=false;
		
		//printf("Pass 1b\n");
		
		//printf("Reg: %i Val: %i\n", pNSF->N163[i].w_length, c_wlength);

		if(c_wlength > 32){
			c_ftnote = FT_NOTE_CUT;
			c_ftoct = 0;
			c_ftinst = 64;
			c_ftpitch = p_ftfx1[ch];
			c_ftwave = p_ftfx2[ch];
			//printf("Goto!\n");
			goto add_n163_item;
		}
		
		//printf("Pass 1c\n");

		if(same_wave){
			for(k=0;k<c_wlength/2;k++){
				if(n163_pwave[i][k]==pNSF->n163_wram[c_waddr+k]) continue;
				same_wave=false;
				break;
			}
		}

		if(same_wave){
			new_wave = false;
			new_inst = false;
			c_ftinst = p_ftinst[ch];
		}
		
		//printf("Pass 2\n");

		if(!same_wave){
			for(j=0;j<n163_count;j++){
				if(n163_addr[j]!=(pNSF->N163[i].w_addr&0x7F)>>1) continue;
				if(n163_length[j]!=4*(64-pNSF->N163[i].w_length)) continue;

				match=true;
				for(k=0;k<n163_length[j]/2;k++){
					if(n163_data[j][k]==pNSF->n163_wram[(n163_addr[j])+k]) continue;
					match=false;
					break;
				}
				if(match){
					c_ftinst = n163_inst[j];
					c_ftwave = n163_wave[j];
					new_inst=false;
					new_wave=false;
					break;
				} 
			}
		}
		
		//printf("Pass 3\n");

		if(new_wave && n163_count<256){	
			for(j=0;j<n163_count;j++){
				int ci = n163_inst[j];
				if(pFTM->Inst[ci].wave_pos!=(pNSF->N163[i].w_addr&0x7F)) continue;
				if(pFTM->Inst[ci].wave_size!=4*(64-pNSF->N163[i].w_length)) continue;
				if(pFTM->Inst[ci].wave_count>=16) continue;

				n163_inst[n163_count] = ci;
				n163_addr[n163_count] = (pNSF->N163[i].w_addr&0x7F)>>1;
				n163_length[n163_count] = 4*(64-pNSF->N163[i].w_length);
				n163_wave[n163_count] = pFTM->Inst[ci].wave_count;
				//for(int k=0;k<n163_length[n163_count]/2;k++){
					
				//}
				memcpy(n163_data[n163_count],&pNSF->n163_wram[n163_addr[n163_count]],n163_length[n163_count]>>1);
				
				pFTM->Inst[ci].wave_count += 1;
				
				c_ftinst = n163_inst[n163_count];
				c_ftwave = n163_wave[n163_count];
				n163_count++;
				new_inst=false;
				new_wave=false;
				break;
			}
		}
		
		//printf("Pass 4\n");

		if(new_inst && n163_count<256 && n163_instruments<16){
			int ci = pFTM->n_instruments;
			n163_inst_idx[n163_instruments] = ci;
			pFTM->Inst[ci].index = ci;
			pFTM->Inst[ci].wave_size=4*(64-pNSF->N163[i].w_length);
			pFTM->Inst[ci].wave_pos=(pNSF->N163[i].w_addr&0x7F);
			pFTM->Inst[ci].type = 5;
			pFTM->Inst[ci].seq_count = 5;		
			pFTM->Inst[ci].wave_count = 1;

			n163_inst[n163_count] = ci;
			n163_addr[n163_count] = (pNSF->N163[i].w_addr&0x7F)>>1;
			n163_length[n163_count] = 4*(64 - pNSF->N163[i].w_length);
			n163_wave[n163_count] = 0;

			memcpy(n163_data[n163_count], &pNSF->n163_wram[n163_addr[n163_count]], n163_length[n163_count]>>1);

			c_ftinst = ci;
			c_ftwave = n163_wave[n163_count];
			
			n163_count++;
			n163_instruments++;
			pFTM->n_instruments++;
		}
		
		//printf("Pass 5\n");

		//if(c_ftinst==64){
		//	printf("FUCK YOU!\n");
		//}

		nes_vol = pNSF->N163[i].vol;
		nes_pitch = pNSF->N163[i].freq;

		Freq = nes_pitch*(2.0f/Channels);
		Register = int(Freq/**8/(64-pNSF->N163[i].w_length)+0.5*/);

		GetN163Note(Register, c_ftnote, c_ftoct, c_ftpitch);
		
		//printf("Pass 6\n");

add_n163_item:

		//printf("Pass 7\n");
		//printf("c_pattern: %i, ch: %i\n", c_pattern, ch);

		Item = &pFTM->Pattern[c_pattern].Item[c_row];

		Item->row		= c_row;
		Item->note       = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftnote;
		Item->octave     = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftoct;
		Item->instrument = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch] && c_ftinst==p_ftinst[ch] && c_ftwave==p_ftfx2[ch])? 64: c_ftinst;
		Item->volume     = (nes_vol==p_ftvol[ch])? 16: nes_vol;
		Item->fxdata[0]  = (c_ftpitch==p_ftfx1[ch])? 0: EF_PITCH;
		Item->fxdata[1]  = (c_ftpitch==p_ftfx1[ch])? 0: c_ftpitch;
		Item->fxdata[2]  = (c_ftwave==p_ftfx2[ch] && c_ftinst==p_ftinst[ch])? 0: EF_DUTY_CYCLE;
		Item->fxdata[3]  = (c_ftwave==p_ftfx2[ch] && c_ftinst==p_ftinst[ch])? 0: c_ftwave;

		p_ftnote[ch] = c_ftnote;
		p_ftoct[ch] = c_ftoct;
		p_ftinst[ch] = c_ftinst;
		p_ftvol[ch] = nes_vol;
		p_ftfx1[ch] = c_ftpitch;
		p_ftfx2[ch] = c_ftwave;
		
		//printf("Pass 8\n");
		//printf("c_waddr: %i c_wlength: %i\n", c_waddr, c_wlength);

		n163_paddr[i] = c_waddr;
		n163_plength[i] = c_wlength;
		memcpy(n163_pwave[i], &pNSF->n163_wram[c_waddr], (c_wlength>>1) & 0xF);
		
		//printf("Pass 9\n");
	}
}

void NSFImporter::Read_FDS(int first_pattern, int c_row)
{
	FTItem* Item;
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;
	int i;
	int ch = fds_ch;

	int c_pattern = first_pattern;

	bool new_wave=true;

	for(int i=0;i<fds_count;i++){
		bool match=true;
		for(int j=0;j<64;j++){
			if(fds_data[i][j]==pNSF->data[0x4040+j]) continue;
			match=false;
			break;
		}
		if(match){
			c_ftinst = fds_inst_idx[i];
			new_wave = false;
			break;
		}
	}

	if(new_wave && fds_count < 16){
		int ci = pFTM->n_instruments;
		fds_inst_idx[fds_count] = ci;
		pFTM->Inst[ci].index = ci;
		pFTM->Inst[ci].type = TYPE_FDS;
		memcpy(fds_data[fds_count],&pNSF->data[0x4040],64);
		memcpy(pFTM->Inst[ci].fds_wave,&pNSF->data[0x4040],64);
		c_ftinst = ci;
		pFTM->n_instruments++;
		fds_count++;
	}

	nes_pitch = ((pNSF->data[0x4083]&0xF)<<8) | pNSF->data[0x4082];
	nes_vol  = (((pNSF->data[0x4080]&0x3F) * 0xF) / 0x20);
    if (nes_vol  > 0xF) nes_vol  = 0xF;
    if (nes_vol  < 0x0) nes_vol  = 0x0;

	GetFDSNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);
	
	Item = &pFTM->Pattern[c_pattern].Item[c_row];

	Item->row		= c_row;
	Item->note       = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch] && c_ftinst==p_ftinst[ch])? 0: c_ftnote;
	Item->octave     = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch] && c_ftinst==p_ftinst[ch])? 0: c_ftoct;
	Item->instrument = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch] && c_ftinst==p_ftinst[ch])? 64: c_ftinst;
	Item->volume     = (nes_vol==p_ftvol[ch])? 16: nes_vol;
	Item->fxdata[0]  = (c_ftpitch==p_ftfx1[ch])? 0: EF_PITCH;
	Item->fxdata[1]  = (c_ftpitch==p_ftfx1[ch])? 0: c_ftpitch;

	p_ftnote[ch] = c_ftnote;
	p_ftoct[ch] = c_ftoct;
	p_ftinst[ch] = c_ftinst;
	p_ftvol[ch] = nes_vol;
	p_ftfx1[ch] = c_ftpitch;
}

void NSFImporter::Read_VRC7(int first_pattern, int c_row)
{
	FTItem* Item;
	for(int i=0;i<6;i++){
		int nes_vol=0;
		int nes_duty=0;
		int nes_pitch=0;
		int c_ftnote=0;
		int c_ftoct=0;
		int c_ftinst = 64;
		int c_ftvol = 16;
		int c_ftpitch=0;
		int c_ftwave=0;
		int c_pattern = first_pattern+i;
		int ch = vrc7_ch + i;
		bool new_vrc7 = true;

		if(pNSF->vrc7_regs[i][2]>>4){
			for(int j=0;j<vrc7_count;j++){
				if(vrc7_patch[j] != pNSF->vrc7_regs[i][2]>>4) continue;
				c_ftinst = vrc7_inst_idx[j];
				new_vrc7 = false;
				break;
			}
		} else {
			for(int j=0;j<vrc7_count;j++){
				bool match=true;
				for(int k=0;k<8;k++){
					if(vrc7_data[j][k] == pNSF->vrc7_custom[k]) continue;
					match = false;
					break;
				}
				if(match){
					c_ftinst = vrc7_inst_idx[j];
					new_vrc7 = false;
					break;
				}
			}
		}

		if(new_vrc7 && vrc7_count<16){
			int ci = pFTM->n_instruments;
			c_ftinst = ci;
			vrc7_inst_idx[vrc7_count] = ci;
			pFTM->Inst[ci].index = ci;
			pFTM->Inst[ci].type = TYPE_VRC7;
			pFTM->Inst[ci].vrc7_patch = pNSF->vrc7_regs[i][2]>>4;
			memcpy(pFTM->Inst[ci].vrc7_data, pNSF->vrc7_custom, 8);			
			vrc7_patch[vrc7_count] = pNSF->vrc7_regs[i][2]>>4;
			memcpy(vrc7_data[vrc7_count], pNSF->vrc7_custom, 8);
			pFTM->n_instruments++;
			vrc7_count++;
		}

		nes_vol = (pNSF->vrc7_regs[i][2]^15)&15;
		nes_pitch = pNSF->N163[i].freq;

		GetVRC7Note(pNSF->vrc7_regs[i][0],pNSF->vrc7_regs[i][1],c_ftnote,c_ftoct,c_ftpitch);

		if(!(pNSF->vrc7_regs[i][1]&0x10)){
			c_ftnote=FT_NOTE_OFF;
			c_ftoct=0;
		}
		
		Item = &pFTM->Pattern[c_pattern].Item[c_row];

		Item->row = c_row;
		Item->note = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftnote;
		Item->octave = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftoct;
		Item->instrument = (c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 64: c_ftinst;
		Item->volume = (nes_vol==p_ftvol[ch])? 16: nes_vol;
		Item->fxdata[0] = (c_ftpitch==p_ftfx1[ch])? 0: EF_PITCH;
		Item->fxdata[1] = (c_ftpitch==p_ftfx1[ch])? 0: c_ftpitch;

		p_ftnote[ch] = c_ftnote;
		p_ftoct[ch] = c_ftoct;
		p_ftvol[ch] = nes_vol;
		p_ftfx1[ch] = c_ftpitch;
	}
}

void NSFImporter::Read_S5B(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;

	for(int i=0;i<3;i++){
		int c_pattern = first_pattern+i;
		int ch = s5b_ch + i;

		nes_vol=pNSF->s5b_regs[8+i]&15;
		nes_pitch=((pNSF->s5b_regs[2*i+1]&15)<<8) | pNSF->s5b_regs[2*i];

		GetNote(nes_pitch,c_ftnote,c_ftoct,c_ftpitch);

		pFTM->Pattern[c_pattern].Item[c_row].row		= c_row;
		pFTM->Pattern[c_pattern].Item[c_row].note       = ((nes_vol==p_ftvol[ch] || i!=0) && c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftnote;
		pFTM->Pattern[c_pattern].Item[c_row].octave     = ((nes_vol==p_ftvol[ch] || i!=0) && c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 0: c_ftoct;
		pFTM->Pattern[c_pattern].Item[c_row].instrument = ((nes_vol==p_ftvol[ch] || i!=0) && c_ftnote==p_ftnote[ch] && c_ftoct==p_ftoct[ch])? 64: s5b_inst;
		pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[ch])? 16: nes_vol;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftfx1[ch])? 0: EF_PITCH;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftfx1[ch])? 0: c_ftpitch;

		p_ftnote[ch]=c_ftnote;
		p_ftoct[ch]=c_ftoct;
		p_ftvol[ch]=nes_vol;
		p_ftfx1[ch]=c_ftpitch;
	}
}


void NSFImporter::GetNote(int Register, int& Note, int& Octave, int& Pitch)
{
    if (Register<=8){
        Note=FT_NOTE_CUT; Octave=0; Pitch=0;
		return;
    }

    int MidiNote = 0;
    int MinDiff = 9999999;
    for (unsigned int i=1; i < 95; i++)
    {
        int Val = freqtable[i];
        int Diff = Val - Register;
        int AbsDiff = (Diff<0)? -Diff: Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            MidiNote = i;
            Pitch = 128+Diff;
        }
    }

    Note = (MidiNote%12)+1;
    Octave = MidiNote/12;
}

void NSFImporter::GetSawNote(int Register, int& Note, int& Octave, int& Pitch)
{
    if (Register<=8)
    {
        Note = FT_NOTE_CUT;
        Octave = 0;
        Pitch = 0;
        return;
    }

    int MidiNote = 0;
    int MinDiff = 9999999;
    for (unsigned int i=1; i < 95; ++i)
    {
        int Val = sawtable[i];
        int Diff = Val - Register;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            MidiNote = i;
            Pitch = 128+Diff;
        }
    }

    Note = (MidiNote % 12) + 1;
    Octave = MidiNote / 12;
}

void NSFImporter::GetN163Note(int Register, int& Note, int& Octave, int& Pitch)
{
	if (Register == 0) {Note = FT_NOTE_CUT; Octave = 0; Pitch = 0; return;}
	int MidiNote=0; int MinDiff=9999999;

	int Channels = pNSF->n163_ch+1;

	for (unsigned int i=1; i<95; ++i){
		int Val = n163table[i];
		int Diff = Register - Val;
		int AbsDiff = (Diff<0) ? -Diff : Diff;
		if (AbsDiff < MinDiff) {MinDiff = AbsDiff; MidiNote = i; Pitch = 128+(Diff/(Channels*2));}
	}
	Note=(MidiNote%12)+1;
	Octave=MidiNote/12;
}

void NSFImporter::GetFDSNote(int Register, int& Note, int& Octave, int& Pitch)
{
	if (Register == 0) {Note = FT_NOTE_CUT; Octave = 0; Pitch = 0; return;}
	int MidiNote=0; int MinDiff=9999999;

    for (unsigned int i=1; i < 95; ++i){
        int Val = fdstable[i];
        int Diff = Register - Val;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff) {MinDiff = AbsDiff; MidiNote = i; Pitch = 128+Diff;}
    }
    Note = (MidiNote % 12) + 1;
    Octave = MidiNote / 12;
}

void NSFImporter::GetVRC7Note(unsigned char RegF, unsigned char RegO, int& Note, int& Octave, int& Pitch)
{
    // octave register high 4 bits ignored
    RegO = RegO & 0x0F;

    if (RegO == 0 && RegF == 0)
    {
        Note = FT_NOTE_CUT;
        Octave = 0;
        Pitch = 0;
        return;
    }

    int Register = RegF + ((RegO & 0x01) << 8); // frequency includes low bit from octave register
    const int FREQ_TABLE[] = {172, 181, 192, 204, 216, 229, 242, 257, 272, 288, 305, 323};

    Octave = RegO >> 1;

    // pitches from below the pitch table should be shifted to an octave we can represent
    while (Register < (FREQ_TABLE[0] - 5) && Octave > 0)
    {
        Register *= 2;
        --Octave;
    }
    // pitches from above the pitch table need to be approximated in a different octave
    while (Register > (FREQ_TABLE[11] + 10) && Octave < 7)
    {
        Register /= 2; // lose one bit of precision, oh well
        ++Octave;
    }

    Note = 0;
    int MinDiff = 9999999;
    for (unsigned int i=0; i < 12; ++i)
    {
        int Val = FREQ_TABLE[i];
        int Diff = Register - Val;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            Note = i + 1;
            Pitch = Diff;
            if (Pitch > 127) Pitch = 127;
            if (Pitch < -128) Pitch = -128;
			Pitch+=128;
        }
    }
}

/*

byte FDS_wave[64] = {0,1,12,22,32,36,39,39,42,47,47,50,48,51,54,58,54,55,49,50,52,61,63,63,59,56,53,51,48,47,41,35,35,35,41,47,48,51,53,56,59,63,63,61,52,50,49,55,54,58,54,51,48,50,47,47,42,39,39,36,32,22,12,1}; 


CNSF_Emu* pNSF = new CNSF_Emu();
CFTMFile* pFTM = new CFTMFile();

int c_sample=0;
int c_inst=1;
int d_count=0;
int d_offset[256]={0};
int d_length[256]={0};
int d_addr[256]={0};
int d_bank[256]={0};

int fds_count=0;
unsigned char fds_data[64][64]={0};

int n163_count=0;
char n163_inst[256]={0};
char n163_wave[256]={0};
char n163_addr[256]={0};
char n163_length[256]={0};
unsigned char n163_data[256][16]={0};

int vrc7_count=0;
char vrc7_patch[64]={0};
unsigned char vrc7_data[64][8]={0};

int n_channels=5;

bool NSF2FTM(CNSF_Emu* pNSF, CFTMFile* pFTM, int length, int track)
{
	int i;//,j,k;
	int c_order=0;
	int c_pattern=0;
	int c_row=0;

	pFTM->ch_fx[0]=2;
	pFTM->ch_fx[1]=2;

	pFTM->expansion=pNSF->Info.expansion & EXPANSION_MASK;

	if(pFTM->expansion & EX_VRC6){
		//pFTM->VRC6_instruments=1;
		//pFTM->Inst_VRC6[0].index=c_inst++;
		pFTM->ch_fx[n_channels]=1;
		pFTM->ch_fx[n_channels+1]=1;
		pFTM->ch_fx[n_channels+2]=1;
		n_channels+=3;
	}

	if(pFTM->expansion & EX_MMC5){
		pFTM->ch_fx[n_channels]=1;
		pFTM->ch_fx[n_channels+1]=1;
		n_channels+=2;
	}

	if(pFTM->expansion & EX_N163){
		//pFTM->N163_instruments=1;
		//pFTM->Inst_N163[0].index=c_inst++;
		for(i=0;i<8;i++) pFTM->ch_fx[n_channels+i]=1;
		n_channels+=8;
	}

	if(pFTM->expansion & EX_FDS){
		//pFTM->FDS_instruments=1;
		//pFTM->Inst_FDS[0].index=c_inst++;
		//memcpy(pFTM->Inst_FDS[0].wave,FDS_wave,64);
		pFTM->ch_fx[n_channels++]=0;
	}

	if(pFTM->expansion & EX_VRC7){
		//pFTM->VRC7_instruments=1;
		//pFTM->Inst_VRC7[0].index=c_inst++;
		for(i=0;i<6;i++) pFTM->ch_fx[n_channels+i]=0;
		n_channels+=6;
	}

	if(pFTM->expansion & EX_S5B){
		pFTM->n_instruments=1;
		pFTM->Inst[0].index=c_inst++;
		for(i=0;i<3;i++) pFTM->ch_fx[n_channels+i]=0;
		n_channels+=3;
	}

	pFTM->channels=n_channels;
	pFTM->N163_channels=8;
	pFTM->n_frames=1;
	pFTM->n_patterns=n_channels*pFTM->n_frames;
	pFTM->speed=1;
	pFTM->tempo=150;
	pFTM->p_length=256;
	pFTM->n_instruments=1;

	if(!pNSF->Init(track)) return false;

	for(i=0;i<length;i++)
	{
		bool w_new = true;
		bool result = pNSF->PlayOnce();
		if(!result) return false;
		c_pattern = (c_order*n_channels);

		NSFImport::Read_2A03(c_pattern,c_row);
		c_pattern+=5;

		if(pFTM->expansion & EX_VRC6){
			NSFImport::Read_VRC6(c_pattern,c_row);
			c_pattern+=3;
		}

		if(pFTM->expansion & EX_MMC5){
			NSFImport::Read_MMC5(c_pattern,c_row);
			c_pattern+=2;
		}

		if(pFTM->expansion & EX_N163){
			NSFImport::Read_N163(c_pattern,c_row);
			c_pattern+=8;
		}

		if(pFTM->expansion & EX_FDS){
			NSFImport::Read_FDS(c_pattern,c_row);
			c_pattern+=1;
		}

		if(pFTM->expansion & EX_VRC7){
			NSFImport::Read_VRC7(c_pattern,c_row);
			c_pattern+=6;
		}

		if(pFTM->expansion & EX_S5B){
			NSFImport::Read_S5B(c_pattern,c_row);
			c_pattern+=3;
		}

		c_row++;

		while(c_row>=256){			
			for(int i=0;i<n_channels;i++){
				c_pattern = (c_order*n_channels)+i;
				pFTM->Pattern[c_pattern].items   = 256;
				pFTM->Pattern[c_pattern].track   = 0;
				pFTM->Pattern[c_pattern].channel = i;
				pFTM->Pattern[c_pattern].index   = c_order;
			}
			//n_items=0;
			c_row-=256;
			c_order++;
			pFTM->n_frames++;
			pFTM->n_patterns += n_channels;
		}
		//pNSF->Clock64();
	}
		
	for(int i=0;i<n_channels;i++){
		c_pattern = (c_order*n_channels)+i;
		pFTM->Pattern[c_pattern].items   = c_row;
		pFTM->Pattern[c_pattern].track   = 0;
		pFTM->Pattern[c_pattern].channel = i;
		pFTM->Pattern[c_pattern].index   = c_order;
	}

	for(int i=0;i<pFTM->n_frames;i++){
		memset(pFTM->fdata[i],i,n_channels);
	}

	pFTM->n_samples=d_count;
	pFTM->Inst[0].type = 1;
	pFTM->n_instruments = 1;

	for(int i=0;i<d_count;i++){
		pFTM->Inst[0].DPCM_data[3*i]=i+1;
		pFTM->Inst[0].DPCM_data[3*i+1]=0xF;
		pFTM->Inst[0].DPCM_data[3*i+2]=0xFF;
		pFTM->Sample[i].index=i;
		pFTM->Sample[i].namelength=0xE;
		pFTM->Sample[i].datalength=d_length[i];
		memcpy(pFTM->Sample[i].name,"New instrument",0xE);
		
		printf("Sample %i bank: %i address: %03X offset: %04X\n",i,d_bank[i],d_addr[i],d_offset[i]);

		if(!(pNSF->is_switched))
			//memcpy(pFTM->Sample[i].data,&pNSF->bdata[d_offset[i]],d_length[i]);
		//else
			memcpy(pFTM->Sample[i].data,&pNSF->data[d_offset[i]],d_length[i]);
	}

	for(int i=0;i<n163_count;i++){
		int ci = n163_inst[i];
		int cw = n163_wave[i];
		for(int j=0;j<n163_length[i]/2;j++){
			pFTM->Inst[ci].wave_data[cw][2*j] = n163_data[i][j]&15;
			pFTM->Inst[ci].wave_data[cw][2*j+1] = n163_data[i][j]>>4;
		}
	}

	return true;
}

bool NSFImport::Import(char* filename,int track, int seconds)
{
	init_tables();

	if(!pNSF->LoadFile(filename)) return false;
	if(!NSF2FTM(pNSF,pFTM,seconds*60,track)) return false;
	if(!pFTM->Optimize())					return false;
	if(!pFTM->SaveFile((char*)"out.ftm")) return false;

	//printf("No errors! Yay. :)\n");
	return true;
}*/