#include <cstring>
#include <math.h>
#include "ft_stuff.h"
#include "NSF_Emu.h"
#include "NSFImport.h"

#define NOTE_COUNT		96
#define MAX_FT_CHANNELS 28

extern CNSF_Emu* pNSF;
extern CFTMFile* pFTM;

extern int c_sample;
extern int c_inst;
extern int d_count;
extern int d_offset[256];
extern int d_length[256];
extern int d_addr[256];
extern int d_bank[256];

extern int fds_count;
extern unsigned char fds_data[64][64];

extern int n163_count;
extern char n163_inst[256];
extern char n163_wave[256];
extern char n163_addr[256];
extern char n163_length[256];
extern unsigned char n163_data[256][16];

char n163_paddr[8];
char n163_plength[8];
char n163_pinst[8];
unsigned char n163_pwave[8][16];

extern int vrc7_count;
extern char vrc7_patch[64];
extern unsigned char vrc7_data[64][8];

unsigned int freqtable[NOTE_COUNT];	// For 2A03/MMC5/VRC6/S5B
unsigned int paltable[NOTE_COUNT];	// For 2A07
unsigned int n163table[NOTE_COUNT];	// For N163
unsigned int sawtable[NOTE_COUNT];	// For VRC6 sawtooth
unsigned int fdstable[NOTE_COUNT];	// For FDS
//unsigned int s5btable[NOTE_COUNT];	// For sunsoft		

int p_ftnote[MAX_FT_CHANNELS]={0};
int p_ftoct[MAX_FT_CHANNELS]={0};
int p_ftinst[MAX_FT_CHANNELS]={64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64};
int p_ftvol[MAX_FT_CHANNELS]={16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16};	
int p_ftduty[MAX_FT_CHANNELS]={0};
int p_ftpitch[MAX_FT_CHANNELS]={0};
int p_ftsweep=0;

void NSFImport::init_tables()
{
	double clock_ntsc = 1789773 / 16.0;
	double clock_pal = 1662607 / 16.0;
	double BASE_FREQ = 32.7032;
	double Freq;
	double Pitch;

	for (int i=0;i<NOTE_COUNT;i++){
		Freq = BASE_FREQ * pow(2.0,i/12.0); // Freq in Hz

		Pitch = (clock_ntsc/Freq)+0.5;
		freqtable[i] = (unsigned int)Pitch;

		Pitch = (clock_pal / Freq) - 0.5;
		paltable[i] = (unsigned int)Pitch;

		Pitch = ((clock_ntsc * 16.0) / (Freq * 14.0)) - 0.5;
		sawtable[i] = (unsigned int)Pitch;

		Pitch = (Freq * 65536.0) / (clock_ntsc / 1.0) + 0.5;
		fdstable[i] = (unsigned int)Pitch;

		Pitch = (Freq * 8 * 983040.0) / clock_ntsc;
		n163table[i] = (unsigned int)(Pitch) / 4;
		if(n163table[i]>0xFFFF) n163table[i]=0xFFFF;
	}
}

void NSFImport::GetNote(int Register, int& Note, int& Octave, int& Pitch)
{
    if (Register<=8){
        Note=FT_NOTE_CUT; Octave=0; Pitch=0;
		return;
    }

    int MidiNote = 0;
    int MinDiff = 9999999;
    for (unsigned int i=1; i < 95; i++)
    {
        int Val = freqtable[i];
        int Diff = Val - Register;
        int AbsDiff = (Diff<0)? -Diff: Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            MidiNote = i;
            Pitch = 128+Diff;
        }
    }

    Note = (MidiNote%12)+1;
    Octave = MidiNote/12;
}

void NSFImport::GetSawNote(int Register, int& Note, int& Octave, int& Pitch)
{
    if (Register<=8)
    {
        Note = FT_NOTE_CUT;
        Octave = 0;
        Pitch = 0;
        return;
    }

    int MidiNote = 0;
    int MinDiff = 9999999;
    for (unsigned int i=1; i < 95; ++i)
    {
        int Val = sawtable[i];
        int Diff = Val - Register;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            MidiNote = i;
            Pitch = 128+Diff;
        }
    }

    Note = (MidiNote % 12) + 1;
    Octave = MidiNote / 12;
}

void NSFImport::GetN163Note(int Register, int& Note, int& Octave, int& Pitch)
{
	if (Register == 0) {Note = FT_NOTE_CUT; Octave = 0; Pitch = 0; return;}
	int MidiNote=0; int MinDiff=9999999;

	int Channels = pNSF->n163_ch+1;

	for (unsigned int i=1; i<95; ++i){
		int Val = n163table[i];
		int Diff = Register - Val;
		int AbsDiff = (Diff<0) ? -Diff : Diff;
		if (AbsDiff < MinDiff) {MinDiff = AbsDiff; MidiNote = i; Pitch = 128+(Diff/(Channels*2));}
	}
	Note=(MidiNote%12)+1;
	Octave=MidiNote/12;
}

void NSFImport::GetFDSNote(int Register, int& Note, int& Octave, int& Pitch)
{
	if (Register == 0) {Note = FT_NOTE_CUT; Octave = 0; Pitch = 0; return;}
	int MidiNote=0; int MinDiff=9999999;

    for (unsigned int i=1; i < 95; ++i){
        int Val = fdstable[i];
        int Diff = Register - Val;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff) {MinDiff = AbsDiff; MidiNote = i; Pitch = 128+Diff;}
    }
    Note = (MidiNote % 12) + 1;
    Octave = MidiNote / 12;
}

void NSFImport::GetVRC7Note(unsigned char RegF, unsigned char RegO, int& Note, int& Octave, int& Pitch)
{
    // octave register high 4 bits ignored
    RegO = RegO & 0x0F;

    if (RegO == 0 && RegF == 0)
    {
        Note = FT_NOTE_CUT;
        Octave = 0;
        Pitch = 0;
        return;
    }

    int Register = RegF + ((RegO & 0x01) << 8); // frequency includes low bit from octave register
    const int FREQ_TABLE[] = {172, 181, 192, 204, 216, 229, 242, 257, 272, 288, 305, 323};

    Octave = RegO >> 1;

    // pitches from below the pitch table should be shifted to an octave we can represent
    while (Register < (FREQ_TABLE[0] - 5) && Octave > 0)
    {
        Register *= 2;
        --Octave;
    }
    // pitches from above the pitch table need to be approximated in a different octave
    while (Register > (FREQ_TABLE[11] + 10) && Octave < 7)
    {
        Register /= 2; // lose one bit of precision, oh well
        ++Octave;
    }

    Note = 0;
    int MinDiff = 9999999;
    for (unsigned int i=0; i < 12; ++i)
    {
        int Val = FREQ_TABLE[i];
        int Diff = Register - Val;
        int AbsDiff = Diff < 0 ? -Diff : Diff;
        if (AbsDiff < MinDiff)
        {
            MinDiff = AbsDiff;
            Note = i + 1;
            Pitch = Diff;
            if (Pitch > 127) Pitch = 127;
            if (Pitch < -128) Pitch = -128;
			Pitch+=128;
        }
    }
}

void NSFImport::Read_2A03(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int nes_sweep=0;
	int dpcm_status=0;
	int dpcm_addr=0;
	int dpcm_length=0;
	int dpcm_earrape=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;
	int c_offset=0;

	int c_pattern = first_pattern;

	nes_vol   = pNSF->data[0x4000]&0xF;
	nes_duty  = pNSF->data[0x4000]>>6;
	nes_pitch = ((pNSF->data[0x4003]&7)<<8) | pNSF->data[0x4002];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[0] && c_ftoct==p_ftoct[0])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[0] && c_ftoct==p_ftoct[0])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[0] && c_ftoct==p_ftoct[0])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[0])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[0])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[0])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[0])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[0])? 0: nes_duty;
	//pFTM->Pattern[c_pattern].Item[c_row].fxdata[4]  = (gb_sweep==p_ftsweep)? 0: (gb_sweep&0x8)? EF_SWEEPUP: EF_SWEEPDOWN;
	//pFTM->Pattern[c_pattern].Item[c_row].fxdata[5]  = (gb_sweep==p_ftsweep && (gb_sweep&0x77))? 0: nes_sweep&0x77;
	
	

	p_ftnote[0]=c_ftnote;
	p_ftoct[0]=c_ftoct;
	p_ftvol[0]=nes_vol;
	p_ftpitch[0]=c_ftpitch;
	p_ftduty[0]=nes_duty;
	//p_ftsweep[0]=nes_sweep;

	c_pattern = first_pattern+1;

	nes_vol   = pNSF->data[0x4004]&0x0F;
	nes_duty  = pNSF->data[0x4004]>>6;
	nes_pitch = ((pNSF->data[0x4007]&7)<<8) | pNSF->data[0x4006];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[1] && c_ftoct==p_ftoct[1])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[1] && c_ftoct==p_ftoct[1])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[1] && c_ftoct==p_ftoct[1])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[1])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[1])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[1])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[1])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[1])? 0: nes_duty;
	//pFTM->Pattern[c_pattern].Item[c_row].fxdata[4]  = (gb_sweep==p_ftsweep)? 0: (gb_sweep&0x8)? EF_SWEEPUP: EF_SWEEPDOWN;
	//pFTM->Pattern[c_pattern].Item[c_row].fxdata[5]  = (gb_sweep==p_ftsweep && (gb_sweep&0x77))? 0: nes_sweep&0x77;

	p_ftnote[1]=c_ftnote;
	p_ftoct[1]=c_ftoct;
	p_ftvol[1]=nes_vol;
	p_ftpitch[1]=c_ftpitch;
	p_ftduty[1]=nes_duty;
	//p_ftsweep[1]=nes_sweep;

	c_pattern = first_pattern+2;

	nes_vol   = (pNSF->data[0x4008]&0x7F) | (pNSF->data[0x400B]>>3);
	nes_pitch = ((pNSF->data[0x400B]&7)<<8) | pNSF->data[0x400A];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[2] && c_ftoct==p_ftoct[2])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[2] && c_ftoct==p_ftoct[2])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[2] && c_ftoct==p_ftoct[2])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[2])? 16: nes_vol? 15: 0;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[2])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[2])? 0: c_ftpitch;

	p_ftnote[2]=c_ftnote;
	p_ftoct[2]=c_ftoct;
	p_ftvol[2]=nes_vol;
	p_ftpitch[2]=c_ftpitch;

	c_pattern = first_pattern+3;

	nes_vol   = pNSF->data[0x400C]&0xF;
	nes_pitch = 31-(pNSF->data[0x400E]&0xF);
	nes_duty = pNSF->data[0x400E]>>7;

	c_ftnote=(nes_pitch%12)+1;
	c_ftoct=nes_pitch/12;

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[3] && c_ftoct==p_ftoct[3])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[3] && c_ftoct==p_ftoct[3])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[3] && c_ftoct==p_ftoct[3])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[3])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (nes_duty==p_ftduty[3])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (nes_duty==p_ftduty[3])? 0: nes_duty;

	p_ftnote[3]=c_ftnote;
	p_ftoct[3]=c_ftoct;
	p_ftvol[3]=nes_vol;
	p_ftduty[3]=nes_duty;

	// DPCM

	c_pattern = first_pattern+4;

	dpcm_status=pNSF->dpcm_triggered;
	pNSF->dpcm_triggered&=0;

	if(dpcm_status){
		bool new_sample=true;

		for(int i=0;i<d_count;i++){
			if(pNSF->is_switched && pNSF->is_FDS){
				int bank=4+(pNSF->data[0x4012]>>6);
				int c_addr=64*(pNSF->data[0x4012]&0x3F);
				c_offset=4096*pNSF->c_bank[bank]+c_addr;
				if(c_offset!=d_offset[i]) continue;
			} else if(pNSF->is_switched && !pNSF->is_FDS){
				int bank=4+(pNSF->data[0x4012]>>6);
				int c_addr=64*(pNSF->data[0x4012]&0x3F);
				c_offset=4096*pNSF->c_bank[bank]+c_addr;
				if(c_offset!=d_offset[i]) continue;			
			} else {
				c_offset=0xC000+64*pNSF->data[0x4012];
				if(c_offset!=d_offset[i]) continue;
			}
			c_sample=i;
			new_sample=false;
			break;
		}

		if(new_sample){
			if(pNSF->is_switched){
				d_bank[d_count]=pNSF->c_bank[4+(pNSF->data[0x4012]>>6)];
				d_addr[d_count]=64*(pNSF->data[0x4012]&0x3F);
				d_offset[d_count]=4096*d_bank[d_count]+d_addr[d_count];
			} else {
				d_bank[d_count]=4+(pNSF->data[0x4012]>>6);
				d_addr[d_count]=64*(pNSF->data[0x4012]&0x3F);
				d_offset[d_count]=0xC000+64*pNSF->data[0x4012];				
			}			
			d_length[d_count]=16*pNSF->data[0x4013]+1;
			if(pNSF->is_switched) memcpy(pFTM->Sample[d_count].data,&pNSF->data[0xC000+64*pNSF->data[0x4012]],d_length[d_count]);
			c_sample=d_count++;
		}
	}

	c_ftpitch = pNSF->data[0x4010]&0xF;
	c_ftnote = dpcm_status? ((c_sample)%12)+1: 0;
	c_ftoct  = dpcm_status? (c_sample)/12: 0;

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[4] && c_ftoct==p_ftoct[4])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[4] && c_ftoct==p_ftoct[4])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (!c_ftnote || c_ftnote==p_ftnote[4] && c_ftoct==p_ftoct[4])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[4])? 0: EF_DPCM_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[4])? 0: c_ftpitch;

	p_ftnote[4]=c_ftnote;
	p_ftoct[4]=c_ftoct;
	p_ftpitch[4]=c_ftpitch;
}

void NSFImport::Read_VRC6(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;

	int c_pattern = first_pattern;

	nes_vol   = pNSF->vrc6_vol[0]&0xF;
	nes_duty  = (pNSF->vrc6_vol[0]&0x70)>>4;
	nes_pitch = pNSF->vrc6_freq[0]&0xFFF;
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	if(!(pNSF->vrc6_freq[0]&0x8000)){
		c_ftnote=FT_NOTE_CUT;
		c_ftoct=0;
		c_ftpitch=p_ftpitch[5];
	}

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[5] && c_ftoct==p_ftoct[5])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[5] && c_ftoct==p_ftoct[5])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[5] && c_ftoct==p_ftoct[5])? 64: pFTM->Inst[0].index;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[5])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[5])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[5])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[5])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[5])? 0: nes_duty;

	p_ftnote[5]=c_ftnote;
	p_ftoct[5]=c_ftoct;
	p_ftvol[5]=nes_vol;
	p_ftpitch[5]=c_ftpitch;
	p_ftduty[5]=nes_duty;

	c_pattern = first_pattern+1;

	nes_vol   = pNSF->vrc6_vol[1]&0xF;
	nes_duty  = (pNSF->vrc6_vol[1]&0x70)>>4;
	nes_pitch = pNSF->vrc6_freq[1]&0xFFF;
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	if(!(pNSF->vrc6_freq[1]&0x8000)){
		c_ftnote=FT_NOTE_CUT;
		c_ftoct=0;
		c_ftpitch=p_ftpitch[6];
	}

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[6] && c_ftoct==p_ftoct[6])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[6] && c_ftoct==p_ftoct[6])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[6] && c_ftoct==p_ftoct[6])? 64: pFTM->Inst[0].index;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[6])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[6])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[6])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[6])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[6])? 0: nes_duty;

	p_ftnote[6]=c_ftnote;
	p_ftoct[6]=c_ftoct;
	p_ftvol[6]=nes_vol;
	p_ftpitch[6]=c_ftpitch;
	p_ftduty[6]=nes_duty;

	c_pattern = first_pattern+2;

	nes_vol   = (pNSF->vrc6_vol[2]&0x1F)>>1;
	nes_duty  = (pNSF->vrc6_vol[2]&0x3F)>>5;
	nes_pitch = pNSF->vrc6_freq[2]&0xFFF;
	GetSawNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	if(!(pNSF->vrc6_freq[2]&0x8000)){
		c_ftnote=FT_NOTE_CUT;
		c_ftoct=0;
		c_ftpitch=p_ftpitch[7];
	}

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[7] && c_ftoct==p_ftoct[7])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[7] && c_ftoct==p_ftoct[7])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[7] && c_ftoct==p_ftoct[7])? 64: pFTM->Inst[0].index;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[7])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[7])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[7])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[7])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[7])? 0: nes_duty;

	p_ftnote[7]=c_ftnote;
	p_ftoct[7]=c_ftoct;
	p_ftvol[7]=nes_vol;
	p_ftpitch[7]=c_ftpitch;
	p_ftduty[7]=nes_duty;
}

void NSFImport::Read_MMC5(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;

	int c_pattern = first_pattern;

	nes_vol   = pNSF->data[0x5000]&0x0F;
	nes_duty  = pNSF->data[0x5000]>>6;
	nes_pitch = ((pNSF->data[0x5003]&7)<<8) | pNSF->data[0x5002];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[8] && c_ftoct==p_ftoct[8])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[8] && c_ftoct==p_ftoct[8])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[8] && c_ftoct==p_ftoct[8])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[8])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[8])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[8])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[8])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[8])? 0: nes_duty;

	p_ftnote[8]=c_ftnote;
	p_ftoct[8]=c_ftoct;
	p_ftvol[8]=nes_vol;
	p_ftpitch[8]=c_ftpitch;
	p_ftduty[8]=nes_duty;

	c_pattern = first_pattern+1;

	nes_vol   = pNSF->data[0x5004]&0x0F;
	nes_duty  = pNSF->data[0x5004]>>6;
	nes_pitch = ((pNSF->data[0x5007]&7)<<8) | pNSF->data[0x5006];
	GetNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		  = c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[9] && c_ftoct==p_ftoct[9])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[9] && c_ftoct==p_ftoct[9])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[9] && c_ftoct==p_ftoct[9])? 64: 0;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[9])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[9])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[9])? 0: c_ftpitch;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (nes_duty==p_ftduty[9])? 0: EF_DUTY_CYCLE;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (nes_duty==p_ftduty[9])? 0: nes_duty;

	p_ftnote[9]=c_ftnote;
	p_ftoct[9]=c_ftoct;
	p_ftvol[9]=nes_vol;
	p_ftpitch[9]=c_ftpitch;
	p_ftduty[9]=nes_duty;
}

void NSFImport::Read_N163(int first_pattern, int c_row)
{
	float Freq;
	int Register;
	int Channels = pNSF->n163_ch+1;

	for(int i=0;i<8;i++){
		int nes_vol=0;
		int nes_duty=0;
		int nes_pitch=0;
		int c_ftnote=0;
		int c_ftoct=0;
		int c_ftinst=0;
		int c_ftvol=0;
		int c_ftpitch=0;
		int c_ftwave=0;
		int c_pattern = first_pattern+i;

		c_ftinst=64;

		bool same_wave=true;
		bool new_wave=true;
		bool new_inst=true;
		int c_n163inst=0;
		int c_wlength = 4*(64-pNSF->N163[i].w_length);
		int c_waddr = (pNSF->N163[i].w_addr&0x7F)>>1;

		if(n163_paddr[i] != c_waddr) same_wave=false;
		if(n163_plength[i] != c_wlength) same_wave=false;

		if(pNSF->N163[i].w_length<56){
			c_ftnote=FT_NOTE_CUT;
			c_ftoct=0;
			c_ftinst=64;
			goto add_n163_item;
		}

		if(same_wave){
			for(int k=0;k<c_wlength/2;k++){
				if(n163_pwave[i][k]==pNSF->n163_wram[c_waddr+k]) continue;
				same_wave=false;
				break;
			}
		}

		if(same_wave){
			new_wave = false;
			new_inst = false;
			c_ftinst = p_ftinst[10+i];
		}

		if(!same_wave){
			for(int j=0;j<n163_count;j++){
				if(n163_addr[j]!=(pNSF->N163[i].w_addr&0x7F)>>1) continue;
				if(n163_length[j]!=4*(64-pNSF->N163[i].w_length)) continue;

				bool match=true;
				for(int k=0;k<n163_length[j]/2;k++){
					if(n163_data[j][k]==pNSF->n163_wram[(n163_addr[j])+k]) continue;
					match=false;
					break;
				}
				if(match){
					c_ftinst=pFTM->Inst[n163_inst[j]].index;
					c_ftwave=n163_wave[j];
					new_inst=false;
					new_wave=false;
					break;
				} 
			}
		}

		if(new_wave && n163_count<256){	
			for(int j=0;j<pFTM->n_instruments;j++){
				if(pFTM->Inst[j].wave_pos!=(pNSF->N163[i].w_addr&0x7F)) continue;
				if(pFTM->Inst[j].wave_size!=4*(64-pNSF->N163[i].w_length)) continue;
				if(pFTM->Inst[j].wave_count==16) continue;

				n163_inst[n163_count]=j;
				n163_addr[n163_count]=(pNSF->N163[i].w_addr&0x7F)>>1;
				n163_length[n163_count]=4*(64-pNSF->N163[i].w_length);
				n163_wave[n163_count]=pFTM->Inst[j].wave_count++;
				//for(int k=0;k<n163_length[n163_count]/2;k++){
					
				//}
				memcpy(n163_data[n163_count],&pNSF->n163_wram[n163_addr[n163_count]],n163_length[n163_count]>>1);

				c_ftinst=pFTM->Inst[n163_inst[n163_count]].index;
				c_ftwave=n163_wave[n163_count];
				n163_count++;
				new_inst=false;
				new_wave=false;
				break;
			}
		}

		if(new_inst && n163_count<256 && pFTM->n_instruments<16){
			pFTM->Inst[pFTM->n_instruments].index=c_inst++;
			pFTM->Inst[pFTM->n_instruments].wave_size=4*(64-pNSF->N163[i].w_length);
			pFTM->Inst[pFTM->n_instruments].wave_pos=(pNSF->N163[i].w_addr&0x7F);

			n163_inst[n163_count]=pFTM->n_instruments;
			n163_addr[n163_count]=(pNSF->N163[i].w_addr&0x7F)>>1;
			n163_length[n163_count]=4*(64-pNSF->N163[i].w_length);
			n163_wave[n163_count]=pFTM->Inst[pFTM->n_instruments].wave_count-1;

			memcpy(n163_data[n163_count],&pNSF->n163_wram[n163_addr[n163_count]],n163_length[n163_count]>>1);

			c_ftinst=pFTM->Inst[pFTM->n_instruments].index;
			c_ftwave=n163_wave[n163_count];
			n163_count++;

			pFTM->n_instruments++;
		}

		//if(c_ftinst==64){
		//	printf("FUCK YOU!\n");
		//}

		nes_vol=pNSF->N163[i].vol;
		nes_pitch=pNSF->N163[i].freq;

		Freq = nes_pitch*(2.0f/Channels);
		Register = int(Freq/**8/(64-pNSF->N163[i].w_length)+0.5*/);

		GetN163Note(Register, c_ftnote, c_ftoct, c_ftpitch);

add_n163_item:

		pFTM->Pattern[c_pattern].Item[c_row].row		= c_row;
		pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[10+i] && c_ftoct==p_ftoct[10+i])? 0: c_ftnote;
		pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[10+i] && c_ftoct==p_ftoct[10+i])? 0: c_ftoct;
		pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[10+i] && c_ftoct==p_ftoct[10+i] && c_ftinst==p_ftinst[10+i] && c_ftwave==p_ftduty[10+i])? 64: c_ftinst;
		pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[10+i])? 16: nes_vol;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[10+i])? 0: EF_PITCH;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[10+i])? 0: c_ftpitch;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[2]  = (c_ftwave==p_ftduty[10+i] && c_ftinst==p_ftinst[10+i])? 0: EF_DUTY_CYCLE;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[3]  = (c_ftwave==p_ftduty[10+i] && c_ftinst==p_ftinst[10+i])? 0: c_ftwave;

		p_ftnote[10+i]=c_ftnote;
		p_ftoct[10+i]=c_ftoct;
		p_ftinst[10+i]=c_ftinst;
		p_ftvol[10+i]=nes_vol;
		p_ftpitch[10+i]=c_ftpitch;
		p_ftduty[10+i]=c_ftwave;

		n163_paddr[i] = c_waddr;
		n163_plength[i] = c_wlength;
		memcpy(n163_pwave[i],&pNSF->n163_wram[c_waddr],c_wlength>>1);
	}
}

void NSFImport::Read_FDS(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;

	int c_pattern = first_pattern;

	bool new_wave=true;

	for(int i=0;i<fds_count;i++){
		bool match=true;
		for(int j=0;j<64;j++){
			if(fds_data[i][j]==pNSF->data[0x4040+j]) continue;
			match=false;
			break;
		}
		if(match){
			c_ftinst=pFTM->Inst[i].index;
			new_wave=false;
			break;
		}
	}

	if(new_wave){
		pFTM->Inst[pFTM->n_instruments].index=c_inst++;
		memcpy(fds_data[fds_count++],&pNSF->data[0x4040],64);
		memcpy(pFTM->Inst[pFTM->n_instruments].fds_wave,&pNSF->data[0x4040],64);
		c_ftinst=pFTM->Inst[pFTM->n_instruments++].index;
	}

	nes_pitch = ((pNSF->data[0x4083]&0xF)<<8) | pNSF->data[0x4082];
	nes_vol  = (((pNSF->data[0x4080]&0x3F) * 0xF) / 0x20);
    if (nes_vol  > 0xF) nes_vol  = 0xF;
    if (nes_vol  < 0x0) nes_vol  = 0x0;

	GetFDSNote(nes_pitch, c_ftnote, c_ftoct, c_ftpitch);

	pFTM->Pattern[c_pattern].Item[c_row].row		= c_row;
	pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[18] && c_ftoct==p_ftoct[18] && c_ftinst==p_ftinst[18])? 0: c_ftnote;
	pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[18] && c_ftoct==p_ftoct[18] && c_ftinst==p_ftinst[18])? 0: c_ftoct;
	pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[18] && c_ftoct==p_ftoct[18] && c_ftinst==p_ftinst[18])? 64: c_ftinst;
	pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[18])? 16: nes_vol;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[18])? 0: EF_PITCH;
	pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[18])? 0: c_ftpitch;

	p_ftnote[18]=c_ftnote;
	p_ftoct[18]=c_ftoct;
	p_ftinst[18]=c_ftinst;
	p_ftvol[18]=nes_vol;
	p_ftpitch[18]=c_ftpitch;
	p_ftduty[18]=nes_duty;
}

void NSFImport::Read_VRC7(int first_pattern, int c_row)
{
	for(int i=0;i<6;i++){
		int nes_vol=0;
		int nes_duty=0;
		int nes_pitch=0;
		int c_ftnote=0;
		int c_ftoct=0;
		int c_ftinst=0;
		int c_ftvol=0;
		int c_ftpitch=0;
		int c_ftwave=0;
		int c_pattern = first_pattern+i;

		bool new_vrc7=true;

		if(pNSF->vrc7_regs[i][2]>>4){
			for(int j=0;j<vrc7_count;j++){
				if(vrc7_patch[j]!=pNSF->vrc7_regs[i][2]>>4) continue;
				c_ftinst=pFTM->Inst[j].index;
				new_vrc7=false;
				break;
			}
		} else {
			for(int j=0;j<vrc7_count;j++){
				bool match=true;
				for(int k=0;k<8;k++){
					if(vrc7_data[j][k]==pNSF->vrc7_custom[k]) continue;
					match=false;
					break;
				}
				if(match){
					c_ftinst=pFTM->Inst[j].index;
					new_vrc7=false;
					break;
				}
			}
		}

		if(new_vrc7 && vrc7_count<59){
			pFTM->Inst[pFTM->n_instruments].index=c_inst++;
			pFTM->Inst[pFTM->n_instruments].vrc7_patch=pNSF->vrc7_regs[i][2]>>4;
			memcpy(pFTM->Inst[pFTM->n_instruments].vrc7_data,pNSF->vrc7_custom,8);
			c_ftinst=pFTM->Inst[pFTM->n_instruments++].index;
			vrc7_patch[vrc7_count]=pNSF->vrc7_regs[i][2]>>4;
			memcpy(vrc7_data[vrc7_count],pNSF->vrc7_custom,8);
			vrc7_count++;			
		}

		nes_vol=(pNSF->vrc7_regs[i][2]^15)&15;
		nes_pitch=pNSF->N163[i].freq;

		GetVRC7Note(pNSF->vrc7_regs[i][0],pNSF->vrc7_regs[i][1],c_ftnote,c_ftoct,c_ftpitch);

		if(!(pNSF->vrc7_regs[i][1]&0x10)){
			c_ftnote=FT_NOTE_OFF;
			c_ftoct=0;
		}

		pFTM->Pattern[c_pattern].Item[c_row].row		= c_row;
		pFTM->Pattern[c_pattern].Item[c_row].note       = (c_ftnote==p_ftnote[19+i] && c_ftoct==p_ftoct[19+i])? 0: c_ftnote;
		pFTM->Pattern[c_pattern].Item[c_row].octave     = (c_ftnote==p_ftnote[19+i] && c_ftoct==p_ftoct[19+i])? 0: c_ftoct;
		pFTM->Pattern[c_pattern].Item[c_row].instrument = (c_ftnote==p_ftnote[19+i] && c_ftoct==p_ftoct[19+i])? 64: c_ftinst;
		pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[19+i])? 16: nes_vol;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[19+i])? 0: EF_PITCH;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[19+i])? 0: c_ftpitch;

		p_ftnote[19+i]=c_ftnote;
		p_ftoct[19+i]=c_ftoct;
		p_ftvol[19+i]=nes_vol;
		p_ftpitch[19+i]=c_ftpitch;
		p_ftduty[19+i]=nes_duty;
	}
}

void NSFImport::Read_S5B(int first_pattern, int c_row)
{
	int nes_vol=0;
	int nes_duty=0;
	int nes_pitch=0;
	int c_ftnote=0;
	int c_ftoct=0;
	int c_ftinst=0;
	int c_ftvol=0;
	int c_ftpitch=0;
	int c_ftwave=0;

	for(int i=0;i<3;i++){
		int c_pattern = first_pattern+i;

		nes_vol=pNSF->s5b_regs[8+i]&15;
		nes_pitch=((pNSF->s5b_regs[2*i+1]&15)<<8) | pNSF->s5b_regs[2*i];

		GetNote(nes_pitch,c_ftnote,c_ftoct,c_ftpitch);

		pFTM->Pattern[c_pattern].Item[c_row].row		= c_row;
		pFTM->Pattern[c_pattern].Item[c_row].note       = ((nes_vol==p_ftvol[25+i] || i!=0) && c_ftnote==p_ftnote[25+i] && c_ftoct==p_ftoct[25+i])? 0: c_ftnote;
		pFTM->Pattern[c_pattern].Item[c_row].octave     = ((nes_vol==p_ftvol[25+i] || i!=0) && c_ftnote==p_ftnote[25+i] && c_ftoct==p_ftoct[25+i])? 0: c_ftoct;
		pFTM->Pattern[c_pattern].Item[c_row].instrument = ((nes_vol==p_ftvol[25+i] || i!=0) && c_ftnote==p_ftnote[25+i] && c_ftoct==p_ftoct[25+i])? 64: pFTM->Inst[0].index;
		pFTM->Pattern[c_pattern].Item[c_row].volume     = (nes_vol==p_ftvol[25+i])? 16: nes_vol;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[0]  = (c_ftpitch==p_ftpitch[25+i])? 0: EF_PITCH;
		pFTM->Pattern[c_pattern].Item[c_row].fxdata[1]  = (c_ftpitch==p_ftpitch[25+i])? 0: c_ftpitch;

		p_ftnote[25+i]=c_ftnote;
		p_ftoct[25+i]=c_ftoct;
		p_ftvol[25+i]=nes_vol;
		p_ftpitch[25+i]=c_ftpitch;
		p_ftduty[25+i]=nes_duty;
	}
}