#define _CRT_SECURE_NO_WARNINGS

#include <cstring>
#include "6502_tables.h"
#include "6502_macros.h"
#include "NSF_Emu.h"

#define PANIC(msg)		\
{						\
	printf("%s\n",msg);	\
	return false;		\
}

bool CNSF_Emu::LogDisasm()
{
	if(!fp_log) PANIC("Couldn't resolve log file! :(\n");
	fprintf(fp_log,"\nNSF DISASM:\n\n");

	for(int i=0x8080;i<0x8088;i++){
		unsigned char c_byte=data[i];
		fprintf(fp_log,"$%04X - %s",i,opcodes[c_byte]);
		for(int j=0;j<opbytes[c_byte]-1;j++){
			fprintf(fp_log," $%02X",data[++i]);
		}
		fprintf(fp_log,"\n");
	}
	return true;
}

bool CNSF_Emu::Init(int track)
{
	fp_log=fopen("NSF_init.log","w"); if(!fp_log) PANIC("Couldn't create log file! :(\n");

	//fprintf(fp_log,"\nNSF INIT LOG:\n\n");

	pc=Info.init_addr;
	a=track;
	x=0;
	y=0;
	flags=4;
	sp=0xFF;

	while(true){
		//fprintf(fp_log,"PC: %04X OP: %02X A:%02X X:%02X Y:%02X F:%02X (%s %s)\n", pc, MEM(pc), a, x, y, flags, opcodes[MEM(pc)],blargg[opmodes[MEM(pc)]]);
		int result=RunCpu();
		if(result==0) break;
		if(result==-1) return false;
	}

	//fprintf(fp_log,"\nInit routine exited properly :)\n\n");
	return true;
}

bool CNSF_Emu::PlayOnce()
{
	pc=Info.play_addr;
	sp=0xFF;

	while(true){
		//fprintf(fp_log,"PC: %04X OP: %02X A:%02X X:%02X Y:%02X F:%02X (%s %s)\n", pc, MEM(pc), a, x, y, flags, opcodes[MEM(pc)],blargg[opmodes[MEM(pc)]]);
		int result=RunCpu();
		//if(result==0) fprintf(fp_log,"\nPlay routine exited properly :)\n\n");
		if(result==0) break;
		if(result==-1) return false;
	}
	//fprintf(fp_log,"SQ1 regs: [%02X,%02X,%02X,%02X]\n",data[0x4000],data[0x4001],data[0x4002],data[0x4003]);
	return true;
}

int CNSF_Emu::RunCpu(void)
{
	unsigned int temp;
	unsigned short temp16;
	unsigned char temp8;
	unsigned char op = MEM(pc); pc++;
	unsigned char temp8a;
	unsigned char temp8b=0;
	unsigned char temp8c=0;
	unsigned char temp8d=0;

	/* printf("PC: %04x OP: %02x A:%02x X:%02x Y:%02x\n", pc-1, op, a, x, y); */

	switch(op)
	{
		case 0xa7: ASSIGNSETFLAGS(a,ABSZP()); x=a; pc++; break;
		case 0xb7: ASSIGNSETFLAGS(a,ABSZPY()); x=a; pc++; break;
		case 0xaf: ASSIGNSETFLAGS(a,ABS16()); x=a; pc+=2; break;
		case 0xa3: ASSIGNSETFLAGS(a,INDX()); x=a; pc++; break;
		case 0xb3: ASSIGNSETFLAGS(a,INDY()); x=a; pc++; break;

		case 0x1a: case 0x3a: case 0x5a:case 0x7a: case 0xda: case 0xfa:
			break;

		case 0x80: case 0x82: case 0x89: case 0xc2: case 0xe2: case 0x04: case 0x44:	case 0x64: case 0x14: case 0x34: case 0x54: case 0x74: case 0xd4: case 0xf4:
			pc++; break;

		case 0x0c: case 0x1c: case 0x3c: case 0x5c: case 0x7c: case 0xdc: case 0xfc:
			pc+=2; break;

		case 0x69: ADC(IMM8()); pc++; break;
		case 0x65: ADC(ABSZP()); pc++; break;
		case 0x75: ADC(ABSZPX()); pc++; break;
		case 0x6d: ADC(ABS16()); pc+=2; break;
		case 0x7d: ADC(ABS16X()); pc+=2; break;
		case 0x79: ADC(ABS16Y()); pc+=2; break;
		case 0x61: ADC(INDX()); pc++; break;
		case 0x71: ADC(INDY()); pc++; break;

		case 0x29: AND(IMM8()); pc++; break;
		case 0x25: AND(ABSZP()); pc++; break;
		case 0x35: AND(ABSZPX()); pc++; break;
		case 0x2d: AND(ABS16()); pc+=2; break;
		case 0x3d: AND(ABS16X()); pc+=2; break;
		case 0x39: AND(ABS16Y()); pc+=2; break;
		case 0x21: AND(INDX()); pc++; break;
		case 0x31: AND(INDY()); pc++; break;

		case 0x0a: ASL(a); break;
		case 0x06: ASL(ABSZP()); pc++; break;
		case 0x16: ASL(ABSZPX()); pc++; break;
		case 0x0e: ASL(ABS16()); pc+=2; break;
		case 0x1e: ASL(ABS16X()); pc+=2; break;

		case 0x90: if (!(flags & FC)) BRANCH() else pc++; break;
		case 0xb0: if (flags & FC) BRANCH() else pc++; break;
		case 0xf0: if (flags & FZ) BRANCH() else pc++; break;

		case 0x24: BIT(ABSZP()); pc++; break;
		case 0x2c: BIT(ABS16()); pc+=2; break;

		case 0x30: if (flags & FN) BRANCH() else pc++; break;
		case 0xd0: if (!(flags & FZ)) BRANCH() else pc++; break;
		case 0x10: if (!(flags & FN)) BRANCH() else pc++; break;
		case 0x50: if (!(flags & FV)) BRANCH() else pc++; break;
		case 0x70: if (flags & FV) BRANCH() else pc++;	break;

		case 0x18: flags &= ~FC; break;
		case 0xd8: flags &= ~FD; break;
		case 0x58: flags &= ~FI; break;
		case 0xb8: flags &= ~FV; break;

		case 0xc9: CMP(a, IMM8()); pc++; break;
		case 0xc5: CMP(a, ABSZP()); pc++; break;
		case 0xd5: CMP(a, ABSZPX()); pc++; break;
		case 0xcd: CMP(a, ABS16()); pc+=2; break;
		case 0xdd: CMP(a, ABS16X()); pc+=2; break;
		case 0xd9: CMP(a, ABS16Y()); pc+=2; break;
		case 0xc1: CMP(a, INDX()); pc++; break;
		case 0xd1: CMP(a, INDY()); pc++; break;
		case 0xe0: CMP(x, IMM8()); pc++; break;
		case 0xe4: CMP(x, ABSZP()); pc++; break;
		case 0xec: CMP(x, ABS16()); pc+=2; break;
		case 0xc0: CMP(y, IMM8()); pc++; break;
		case 0xc4: CMP(y, ABSZP()); pc++; break;
		case 0xcc: CMP(y, ABS16()); pc+=2; break;

		case 0xc6: DEC(ABSZP()); WRITE(IMM8(),ABSZP()); pc++; break;
		case 0xd6: DEC(ABSZPX()); WRITE(IMM8X(),ABSZPX()); pc++; break;
		case 0xce: DEC(ABS16()); WRITE(IMM16(),ABS16()); pc+=2; break;
		case 0xde: DEC(ABS16X()); WRITE(IMM16X(),ABS16X()); pc+=2;	break;

		case 0xca: x--;SETFLAGS(x); break;
		case 0x88: y--;SETFLAGS(y); break;

		case 0x49: EOR(IMM8()); pc++; break;
		case 0x45: EOR(ABSZP()); pc++; break;
		case 0x55: EOR(ABSZPX()); pc++; break;
		case 0x4d: EOR(ABS16()); pc+=2; break;
		case 0x5d: EOR(ABS16X()); pc+=2; break;
		case 0x59: EOR(ABS16Y()); pc+=2; break;
		case 0x41: EOR(INDX()); pc++; break;
		case 0x51: EOR(INDY()); pc++; break;

		case 0xe6: INC(ABSZP()); WRITE(IMM8(),ABSZP()); pc++; break;
		case 0xf6: INC(ABSZPX()); WRITE(IMM8X(),ABSZPX()); pc++; break;
		case 0xee: INC(ABS16()); WRITE(IMM16(),ABS16()); pc+=2; break;
		case 0xfe: INC(ABS16X()); WRITE(IMM16X(),ABS16X()); pc+=2; break;

		case 0xe8: x++; SETFLAGS(x); break;
		case 0xc8: y++; SETFLAGS(y); break;

		case 0x20:
			PUSH((pc+1)>>8);
			PUSH((pc+1)&0xff);
			pc = ABSOLUTE();
			break;

		case 0x4c: pc = ABSOLUTE(); break;

		case 0x6c: {
				unsigned short adr = ABSOLUTE();
				pc = (MEM(adr) | (MEM(((adr + 1) & 0xff) | (adr & 0xff00)) << 8));
			}
			break;

		case 0xa9: ASSIGNSETFLAGS(a,IMM8()); pc++; break;
		case 0xa5: ASSIGNSETFLAGS(a,ABSZP()); pc++; break;
		case 0xb5: ASSIGNSETFLAGS(a,ABSZPX()); pc++; break;
		case 0xad: ASSIGNSETFLAGS(a,ABS16()); READ(IMM16()); pc+=2; break;
		case 0xbd: ASSIGNSETFLAGS(a,ABS16X()); pc+=2; break;
		case 0xb9: ASSIGNSETFLAGS(a,ABS16Y()); pc+=2; break;
		case 0xa1: ASSIGNSETFLAGS(a,INDX()); pc++; break;
		case 0xb1: ASSIGNSETFLAGS(a,INDY()); pc++; break;
		case 0xa2: ASSIGNSETFLAGS(x,IMM8()); pc++; break;
		case 0xa6: ASSIGNSETFLAGS(x,ABSZP()); pc++; break;
		case 0xb6: ASSIGNSETFLAGS(x,ABSZPY()); pc++; break;
		case 0xae: ASSIGNSETFLAGS(x,ABS16()); READ(IMM16()); pc+=2; break;
		case 0xbe: ASSIGNSETFLAGS(x,ABS16Y()); pc+=2; break;
		case 0xa0: ASSIGNSETFLAGS(y,IMM8()); pc++; break;
		case 0xa4: ASSIGNSETFLAGS(y,ABSZP()); pc++; break;
		case 0xb4: ASSIGNSETFLAGS(y,ABSZPX()); pc++; break;
		case 0xac: ASSIGNSETFLAGS(y,ABS16()); READ(IMM16()); pc+=2; break;
		case 0xbc: ASSIGNSETFLAGS(y,ABS16X()); pc+=2; break;

		case 0x4a: LSR(a); break;
		case 0x46: LSR(ABSZP()); WRITE(IMM8(),ABSZP()); pc++; break;
		case 0x56: LSR(ABSZPX()); WRITE(IMM8X(),ABSZPX()); pc++; break;
		case 0x4e: LSR(ABS16()); WRITE(IMM16(),ABS16()); pc+=2; break;
		case 0x5e: LSR(ABS16X()); WRITE(IMM16X(),ABS16X()); pc+=2; break;

		case 0xea: break;

		case 0x09: ORA(IMM8()); pc++; break;
		case 0x05: ORA(ABSZP()); pc++; break;
		case 0x15: ORA(ABSZPX()); pc++; break;
		case 0x0d: ORA(ABS16()); pc+=2; break;
		case 0x1d: ORA(ABS16X()); pc+=2; break;
		case 0x19: ORA(ABS16Y()); pc+=2; break;
		case 0x01: ORA(INDX()); pc++; break;
		case 0x11: ORA(INDY()); pc++; break;

		case 0x48: PUSH(a); break;
		case 0x08: PUSH(flags); break;
		case 0x68: ASSIGNSETFLAGS(a, POP()); break;
		case 0x28: flags = POP(); break;

		case 0x2a: ROL(a); break;
		case 0x26: ROL(ABSZP()); WRITE(IMM8(),ABSZP()); pc++; break;
		case 0x36: ROL(ABSZPX()); WRITE(IMM8X(),ABSZPX()); pc++; break;
		case 0x2e: ROL(ABS16()); WRITE(IMM16(),ABS16()); pc+=2; break;
		case 0x3e: ROL(ABS16X()); WRITE(IMM16X(),ABS16X()); pc+=2; break;
		case 0x6a: ROR(a); break;
		case 0x66: ROR(ABSZP()); WRITE(IMM8(),ABSZP()); pc++; break;
		case 0x76: ROR(ABSZPX()); WRITE(IMM8X(),ABSZPX()); pc++; break;
		case 0x6e: ROR(ABS16()); WRITE(IMM16(),ABS16()); pc+=2; break;
		case 0x7e: ROR(ABS16X()); WRITE(IMM16X(),ABS16X()); pc+=2; break;

		case 0x40:
			if (sp == 0xff) return 0;
			flags = POP();
			pc = POP();
			pc |= POP() << 8;
		break;

		case 0x60:
			if (sp == 0xff) return 0;
			pc = POP();
			pc |= POP() << 8;
			pc++;
		break;

		case 0xe9: SBC(IMM8()); pc++; break;
		case 0xe5: SBC(ABSZP()); pc++; break;
		case 0xf5: SBC(ABSZPX()); pc++; break;
		case 0xed: SBC(ABS16()); pc+=2; break;
		case 0xfd: SBC(ABS16X()); pc+=2; break;
		case 0xf9: SBC(ABS16Y()); pc+=2; break;
		case 0xe1: SBC(INDX()); pc++; break;
		case 0xf1: SBC(INDY()); pc++; break;

		case 0x38: flags |= FC; break;
		case 0xf8: flags |= FD; break;
		case 0x78: flags |= FI; break;

		case 0x85: WRITE(IMM8(),a); pc++; break;
		case 0x95: WRITE(IMM8X(),a); pc++; break;
		case 0x8d: WRITE(IMM16(),a); pc+=2; break;
		case 0x9d: WRITE(IMM16X(),a); pc+=2; break;
		case 0x99: WRITE(IMM16Y(),a); pc+=2; break;
		case 0x81: WRITE(IDDRX(),a); pc++; break;
		case 0x91: WRITE(IDDRY(),a); pc++; break;

		case 0x86: WRITE(IMM8(),x); pc++; break;
		case 0x96: WRITE(IMM8Y(),x); pc++; break;
		case 0x8e: WRITE(IMM16(),x); pc+=2; break;

		case 0x84: WRITE(IMM8(),y); pc++; break;
		case 0x94: WRITE(IMM8X(),y); pc++; break;
		case 0x8c: WRITE(IMM16(),y); pc+=2; break;

		case 0xaa: ASSIGNSETFLAGS(x, a); break;
		case 0xba: ASSIGNSETFLAGS(x, sp); break;
		case 0x8a: ASSIGNSETFLAGS(a, x); break;
		case 0x9a: ASSIGNSETFLAGS(sp, x); break;
		case 0x98: ASSIGNSETFLAGS(a, y); break;
		case 0xa8: ASSIGNSETFLAGS(y, a); break;

		case 0x00: return 0;

		case 0x02:
			printf("Error: CPU halt at %04X\n", pc-1);
			return -1;

		default:
			printf("Error: Unknown opcode $%02X at $%04X\n", op, pc-1);
			return -1;
	}
	return 1;
}
