#define _CRT_SECURE_NO_WARNINGS

#include <cstdio>
#include <cstring>
#include "FTMFile.h"
#include "NSF_Emu.h"

#define NSF_HEADER_SIZE 0x80
#define MEM_SIZE		65536

#define PANIC(msg)		\
{						\
	printf("%s\n",msg);	\
	return false;		\
}						\

CNSF_Emu::CNSF_Emu()
{
	fp_src=NULL;
	fp_dst=NULL;
	fp_log=NULL;
	src_size=0;
	dpcm_triggered=0;
	data=new unsigned char[MEM_SIZE];
	op_count=new int[256];
	memset(&Info,0,sizeof(stInfo));
	memset(data,0,MEM_SIZE);
	memset(op_count,0,256*sizeof(int));
	memset(N163,0,8*sizeof(stRegs_N163));
	is_switched=false;
	is_FDS=false;
	n163_addr=0;
	vrc7_addr=0;
	s5b_addr=0;
}

bool CNSF_Emu::LoadFile(char* filename)
{
	fp_src=fopen(filename,"rb"); if(!fp_src) PANIC("Couldn't open/find source file! :(\n");
	fp_log=fopen("NSF_load.log","w"); if(!fp_log) PANIC("Couldn't create log file! :(\n");

	fseek(fp_src,0,SEEK_END); src_size=ftell(fp_src); fseek(fp_src,0,SEEK_SET);

	if(!ReadHeader()) return false;
	if(!ReadData()) return false;

	//if(!LogDisasm()) return false;

	fclose(fp_src);
	fclose(fp_log);
	fp_src=NULL;
	fp_log=NULL;
	return true;
}

bool CNSF_Emu::ReadHeader()
{
	if(src_size<NSF_HEADER_SIZE) PANIC("NSF file too small for even just a header! :(");
	fread(&Info,NSF_HEADER_SIZE,1,fp_src);
	if(Info.expansion&0x04) is_FDS=true;
	for(int i=0;i<8;i++){
		c_bank[i]=Info.banks_init[i];
		if(Info.banks_init[i]) is_switched=true;
	}
	return LogHeader();
}

bool CNSF_Emu::ReadData()
{
	int read_size=src_size-ftell(fp_src);
	int load_offset=Info.load_addr-0x8000;
	if(load_offset<0) PANIC("Unknown data format! :(");
	if(is_switched && !is_FDS){
		n_banks=(load_offset+read_size+0x0FFF)>>12;
		bdata=new byte[n_banks<<12];
		memset(bdata,0,n_banks<<12);
		fread(bdata+load_offset,1,read_size,fp_src);
		for(int i=0;i<8;i++){
			data[0x5FF8+i]=Info.banks_init[i];
			if(Info.banks_init[i]>=n_banks) continue;
			memcpy(&data[0x8000+4096*i],&bdata[4096*Info.banks_init[i]],0x1000);
		}
	} else if(is_switched && is_FDS){
		n_banks=(load_offset+read_size+0x0FFF)>>12;
		bdata=new byte[n_banks<<12];
		memset(bdata,0,n_banks<<12);
		fread(bdata+load_offset,1,read_size,fp_src);
		data[0x5FF6]=Info.banks_init[6];
		data[0x5FF7]=Info.banks_init[7];
		memcpy(&data[0x6000],&bdata[4096*Info.banks_init[6]],0x1000);
		memcpy(&data[0x7000],&bdata[4096*Info.banks_init[7]],0x1000);
		for(int i=0;i<8;i++){
			data[0x5FF8+i]=Info.banks_init[i];
			if(Info.banks_init[i]>=n_banks) continue;
			memcpy(&data[0x8000+4096*i],&bdata[4096*Info.banks_init[i]],0x1000);
		}
	} else {
		fread(&data[Info.load_addr],read_size,1,fp_src);
	}


	return LogData();
}

bool CNSF_Emu::LogHeader()
{
	if(!fp_log) PANIC("Couldn't resolve log file! :(\n");

	fprintf(fp_log,"NSF HEADER:\n\n");
	fprintf(fp_log,"Version number: %i\n",Info.version);
	fprintf(fp_log,"Number of songs: %i\n",Info.n_songs);
	fprintf(fp_log,"First song: %i\n",Info.first);
	fprintf(fp_log,"Load address: $%04X\n",Info.load_addr);
	fprintf(fp_log,"Init address: $%04X\n",Info.init_addr);
	fprintf(fp_log,"Play address: $%04X\n",Info.play_addr);
	fprintf(fp_log,"NTSC speed: $%04X\n",Info.ntsc_speed);
	fprintf(fp_log,"PAL speed: $%04X\n",Info.pal_speed);

	fprintf(fp_log,"Bankswitch init values:");
	for(int i=0;i<8;i++) fprintf(fp_log," $%02X",Info.banks_init[i]);
	fprintf(fp_log,"\n");

	fprintf(fp_log,"Title: %s\n",Info.title);
	fprintf(fp_log,"Author: %s\n",Info.author);
	fprintf(fp_log,"Copyright: %s\n",Info.ccccc);

	return true;
}

bool CNSF_Emu::LogData()
{
	if(!fp_log) PANIC("Couldn't resolve log file! :(\n");
	fprintf(fp_log,"\nNSF DATA:\n\n");
	fprintf(fp_log,"Data size: %i\n\n",src_size-NSF_HEADER_SIZE);
	return true;
}
