/*
** FamiTracker - NES/Famicom sound tracker
** Copyright (C) 2005-2012  Jonathan Liss
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful, 
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
** Library General Public License for more details.  To obtain a 
** copy of the GNU Library General Public License, write to the Free 
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
*/

// Part of this is directly taken from FamiTracker's source, hence the copyright notice. ;)
// -Slimeball

typedef unsigned char byte;

#define EX_NONE		0x00
#define EX_VRC6		0x01
#define EX_VRC7		0x02
#define EX_FDS		0x04
#define EX_MMC5		0x08
#define EX_N163		0x10
#define EX_S5B		0x20

#define FT_NOTE_OFF	0x0D
#define FT_NOTE_CUT	0x0E

enum{
	TYPE_2A03=1,
	TYPE_VRC6,
	TYPE_VRC7,
	TYPE_FDS,
	TYPE_N163,
	TYPE_S5B
};


#define CHANNELS_2A03 5
#define CHANNELS_MMC5 7
#define CHANNELS_VRC6 8
#define CHANNELS_VRC7 11
#define CHANNELS_FDS  6
#define CHANNELS_N163 13

#define MAX_I_SIZE_2A03		11
#define MAX_I_SIZE_MMC5		11
#define MAX_I_SIZE_VRC6		11
#define MAX_I_SIZE_VRC7		11
#define MAX_I_SIZE_FDS		28
#define MAX_I_SIZE_N163		125
#define MAX_I_SIZE_N163_OLD 125

//byte ft_driver[16]={0x4C,0x33,0x80,0x4C,0x97,0x82,0xBD,0x2E,0x80,0x49,0xFF,0x2D,0x09,0x02,0x8D,0x09};

#define max(a,b)            (((a) > (b)) ? (a) : (b))
#define min(a,b)            (((a) < (b)) ? (a) : (b))

//Versions
enum{
	FT_2_7BETA1,
	FT_2_7BETA2,
	FT_2_7BETA3,
	FT_2_7,
	FT_2_8BETA,
	FT_2_9,
	FT_3_0BETA1,
	FT_3_0BETA1A,
	FT_3_0BETA2,
	FT_3_0,
	FT_3_1BETA1,
	FT_3_1BETA2,
	FT_3_1BETA3,
	FT_3_1BETA4,
	FT_3_5,
	FT_3_6BETA1,
	FT_3_6BETA2,
	FT_3_6,
	FT_3_7BETA1,
	FT_3_7BETA2,
	FT_3_7,
	FT_3_8BETA1,
	FT_3_8BETA2,
	FT_4_0,
	FT_4_1,
	FT_4_2BETA,
	FT_4_2,
	FT_4_3BETA1,
	FT_4_3BETA1A,
	FT_4_3BETA2,
	FT_4_3BETA3,
	FT_VERSIONS
};

// Channel effects
#define DEF_CMD(x) ((x << 1) | 0x80)
enum {
	EF_NONE = 0,
	EF_SPEED,
	EF_JUMP,
	EF_SKIP,
	EF_HALT,
	EF_VOLUME,
	EF_PORTAMENTO,
	EF_PORTAOFF,				// unused!!
	EF_SWEEPUP,
	EF_SWEEPDOWN,
	EF_ARPEGGIO,
	EF_VIBRATO,
	EF_TREMOLO,
	EF_PITCH,
	EF_DELAY,
	EF_EARRAPE,
	EF_PORTA_UP,
	EF_PORTA_DOWN,
	EF_DUTY_CYCLE,
	EF_SAMPLE_OFFSET,
	EF_SLIDE_UP,
	EF_SLIDE_DOWN,
	EF_VOLUME_SLIDE,
	EF_NOTE_CUT,
	EF_RETRIGGER,
	EF_DELAYED_VOLUME,			// Unimplemented
	EF_FDS_MOD_DEPTH,
	EF_FDS_MOD_SPEED_HI,
	EF_FDS_MOD_SPEED_LO,
	EF_DPCM_PITCH,
	EF_SUNSOFT_ENV_LO,
	EF_SUNSOFT_ENV_HI,
	EF_SUNSOFT_ENV_TYPE,
//	EF_TARGET_VOLUME_SLIDE, 
/*
	EF_VRC7_MODULATOR,
	EF_VRC7_CARRIER,
	EF_VRC7_LEVELS,
*/
	EF_COUNT
};

// Command table
const unsigned char CMD_INSTRUMENT			= DEF_CMD(0);	// 80h
const unsigned char CMD_EFF_SPEED			= DEF_CMD(1);	// 82h
const unsigned char CMD_EFF_JUMP			= DEF_CMD(2);	// ...
const unsigned char CMD_EFF_SKIP			= DEF_CMD(3);
const unsigned char CMD_EFF_HALT			= DEF_CMD(4);
const unsigned char CMD_EFF_VOLUME			= DEF_CMD(5);
const unsigned char CMD_EFF_PORTAMENTO		= DEF_CMD(6);
const unsigned char CMD_EFF_PORTAUP			= DEF_CMD(7);
const unsigned char CMD_EFF_PORTADOWN		= DEF_CMD(8);
const unsigned char CMD_EFF_SWEEP			= DEF_CMD(9);
const unsigned char CMD_EFF_ARPEGGIO		= DEF_CMD(10);
const unsigned char CMD_EFF_VIBRATO			= DEF_CMD(11);
const unsigned char CMD_EFF_TREMOLO			= DEF_CMD(12);
const unsigned char CMD_EFF_PITCH			= DEF_CMD(13);
const unsigned char CMD_EFF_DELAY			= DEF_CMD(14);
const unsigned char CMD_EFF_EARRAPE			= DEF_CMD(15);
const unsigned char CMD_EFF_DUTY			= DEF_CMD(16);
const unsigned char CMD_EFF_OFFSET			= DEF_CMD(17);
const unsigned char CMD_EFF_SLIDE_UP		= DEF_CMD(18);
const unsigned char CMD_EFF_SLIDE_DOWN		= DEF_CMD(19);
const unsigned char CMD_EFF_VOL_SLIDE		= DEF_CMD(20);
const unsigned char CMD_EFF_NOTE_CUT		= DEF_CMD(21);
const unsigned char CMD_EFF_RETRIGGER		= DEF_CMD(22);
const unsigned char CMD_EFF_DPCM_PITCH		= DEF_CMD(23);

const unsigned char CMD_SET_DURATION		= DEF_CMD(24);	// B0h (change these in init.s)
const unsigned char CMD_RESET_DURATION		= DEF_CMD(25);	// B2h

const unsigned char CMD_EFF_FDS_MOD_DEPTH	= DEF_CMD(26);
const unsigned char CMD_EFF_FDS_MOD_RATE_HI = DEF_CMD(27);
const unsigned char CMD_EFF_FDS_MOD_RATE_LO = DEF_CMD(28);

const unsigned char CMD_EFF_VRC7_PATCH		= DEF_CMD(26);	// TODO: hack, fix this

const unsigned char CMD_LOOP_POINT			= DEF_CMD(26);	// Currently not in use

/*
	EF_SPEED,	EF_JUMP,	EF_SKIP,	EF_HALT,	EF_VOLUME,	EF_PORTAMENTO,
	EF_PORTAOFF,				// unused!!
	EF_SWEEPUP,	EF_SWEEPDOWN,	EF_ARPEGGIO,	EF_VIBRATO,	EF_TREMOLO,	EF_PITCH,
	EF_DELAY,	EF_EARRAPE,	EF_PORTA_UP,	EF_PORTA_DOWN,	EF_DUTY_CYCLE,	EF_SAMPLE_OFFSET,
	EF_SLIDE_UP,	EF_SLIDE_DOWN,	EF_VOLUME_SLIDE,	EF_NOTE_CUT,	EF_RETRIGGER,	EF_DELAYED_VOLUME,			// Unimplemented
	EF_FDS_MOD_DEPTH,	EF_FDS_MOD_SPEED_HI,	EF_FDS_MOD_SPEED_LO,	EF_DPCM_PITCH,	EF_SUNSOFT_ENV_LO,
	EF_SUNSOFT_ENV_HI,	EF_SUNSOFT_ENV_TYPE,
	*/
inline byte convert_cmd(byte cmd){
	if(cmd==0x99) return 99;
	if(cmd==0x9F) return 99;
	if(cmd==0x87) return 99;
	if(cmd==0x8D) return 99;
	if(cmd==CMD_INSTRUMENT){return 100;}
	if(cmd==CMD_EFF_SPEED){return EF_SPEED;}
	if(cmd==CMD_EFF_JUMP){return EF_JUMP;}
	if(cmd==CMD_EFF_SKIP){return EF_SKIP;}
	if(cmd==CMD_EFF_HALT){return EF_HALT;}
	if(cmd==CMD_EFF_VOLUME){return EF_VOLUME;}
	if(cmd==CMD_EFF_PORTAMENTO){return EF_PORTAMENTO;}
	if(cmd==CMD_EFF_PORTAUP){return EF_PORTA_UP;}
	if(cmd==CMD_EFF_PORTADOWN){return EF_PORTA_DOWN;}
	if(cmd==CMD_EFF_SWEEP){return 101;}
	if(cmd==CMD_EFF_ARPEGGIO){return EF_ARPEGGIO;}
	if(cmd==CMD_EFF_VIBRATO){return EF_VIBRATO;}
	if(cmd==CMD_EFF_TREMOLO){return EF_TREMOLO;}
	if(cmd==CMD_EFF_PITCH){return EF_PITCH;}
	if(cmd==CMD_EFF_DELAY){return EF_DELAY;}
	if(cmd==CMD_EFF_EARRAPE){return EF_EARRAPE;}
	if(cmd==CMD_EFF_DUTY){return EF_DUTY_CYCLE;}
	if(cmd==CMD_EFF_OFFSET){return EF_SAMPLE_OFFSET;}
	if(cmd==CMD_EFF_SLIDE_UP){return EF_SLIDE_UP;}
	if(cmd==CMD_EFF_SLIDE_DOWN){return EF_SLIDE_DOWN;}
	if(cmd==CMD_EFF_VOL_SLIDE){return EF_VOLUME_SLIDE;}
	if(cmd==CMD_EFF_NOTE_CUT){return EF_NOTE_CUT;}
	if(cmd==CMD_EFF_RETRIGGER){return EF_RETRIGGER;}
	if(cmd==CMD_EFF_DPCM_PITCH){return EF_DPCM_PITCH;}
	if(cmd==CMD_EFF_FDS_MOD_DEPTH){return EF_FDS_MOD_DEPTH;}
	if(cmd==CMD_EFF_FDS_MOD_RATE_HI){return EF_FDS_MOD_SPEED_HI;}
	if(cmd==CMD_EFF_FDS_MOD_RATE_LO){return EF_FDS_MOD_SPEED_LO;}
	if(cmd==CMD_EFF_VRC7_PATCH){return 102;}
	if(cmd==CMD_LOOP_POINT){return 102;}
	return 103;
}

inline char ftmn2not(byte ftmn)
{
	if(ftmn==0x00) return '-';
	if(ftmn==0x13) return '=';
	if(ftmn==0x14) return '^';
	switch((ftmn+1)%12){
	case  0: case  1: return 'C';
	case  2: case  3: return 'D';
	case  4: case  5: return 'E';
	case  6:		  return 'F';
	case  7: case  8: return 'G';
	case  9: case 10: return 'A';
	case 11:		  return 'B';
	default:		  return 'X';
	}
}

inline char ftmn2acc(byte ftmn)
{
	if(ftmn==0x00) return '-';
	if(ftmn==0x13) return '=';
	if(ftmn==0x14) return '^';
	switch((ftmn+1)%12){
	case  1: return '#';
	case  3: return '#';
	case  6: return '#';
	case  8: return '#';
	case 10: return '#';
	default: return '-';
	}	
}

inline char ftmno2oct(byte ftmn, byte ftmo)
{
	if(ftmn==0x00) return '-';
	if(ftmn==0x13) return '=';
	if(ftmn==0x14) return '^';;
	return 48+ftmo;
}

inline char ftmfx(byte rawfx)
{
	switch(rawfx){
	case 0:					return '-';
	case EF_SPEED:			return 'F';
	case EF_JUMP:			return 'B';
	case EF_SKIP:			return 'D';
	case EF_HALT:			return 'C';
	case EF_VOLUME:			return 'E';
	case EF_PORTAMENTO:		return '3';
	case EF_PORTA_UP:		return '1';
	case EF_PORTA_DOWN:		return '2';
	case EF_SWEEPUP:		return 'H';
	case EF_SWEEPDOWN:		return 'I';
	case EF_ARPEGGIO:		return '0';
	case EF_VIBRATO:		return '4';
	case EF_TREMOLO:		return '7';
	case EF_PITCH:			return 'P';
	case EF_DELAY:			return 'G';
	case EF_EARRAPE:		return 'Z';
	case EF_DUTY_CYCLE:		return 'V';
	case EF_SAMPLE_OFFSET:	return 'Y';
	case EF_SLIDE_UP:		return 'Q';
	case EF_SLIDE_DOWN:		return 'R';
	case EF_VOLUME_SLIDE:	return 'A';
	case EF_NOTE_CUT:		return 'S';
	case EF_RETRIGGER:		return 'X';
	case EF_DPCM_PITCH:		return 'W';
	default: return '?';
	}	
}